"""Constants for the executors."""
from pathlib import Path

SECRETS_VOLUME = Path("/secrets")
SOCKETS_VOLUME = Path("/sockets")
INPUTS_VOLUME = Path("/inputs")
PROCESSING_VOLUME = Path("/processing")

CONTAINER_TIMEOUT = 600

# Relative path to socket files inside sockets volume.
COMMUNICATION_PROCESSING_SOCKET = "_socket1.s"
SCRIPT_SOCKET = "_socket2.s"
UPLOAD_FILE_SOCKET = "_upload_socket.s"

TMPDIR = ".tmp"

INPUTS_VOLUME_NAME = "FLOW_INPUT_VOLUME"
PROCESSING_VOLUME_NAME = "FLOW_PROCESSING_VOLUME"



# pylint: disable=missing-docstring
from datetime import timedelta
from unittest.mock import MagicMock, patch

from django.db.utils import IntegrityError
from django.utils import timezone

from resolwe.storage.connectors import (
    AwsS3Connector,
    GoogleConnector,
    LocalFilesystemConnector,
)
from resolwe.storage.models import (
    AccessLog,
    FileStorage,
    ReferencedPath,
    StorageLocation,
)
from resolwe.test import TransactionTestCase

CONNECTORS_SETTINGS = {
    "local": {
        "connector": "resolwe.storage.connectors.localconnector.LocalFilesystemConnector",
        "config": {
            "priority": 1,
            "path": "/",
            "delete": {
                "delay": 2,
                "min_other_copies": 2,
            },
        },
    },
    "S3": {
        "connector": "resolwe.storage.connectors.s3connector.AwsS3Connector",
        "config": {
            "priority": 200,
            "bucket": "genialis-test-storage",
            "credentials": "test.json",
        },
    },
    "GCS": {
        "connector": "resolwe.storage.connectors.googleconnector.GoogleConnector",
        "config": {
            "bucket": "genialis-test-storage",
            "credentials": "test.json",
        },
    },
}

CONNECTORS = {
    "local": LocalFilesystemConnector(CONNECTORS_SETTINGS["local"]["config"], "local"),
    "GCS": GoogleConnector(CONNECTORS_SETTINGS["GCS"]["config"], "GCS"),
    "S3": AwsS3Connector(CONNECTORS_SETTINGS["S3"]["config"], "S3"),
}


@patch("resolwe.storage.models.connectors", CONNECTORS)
class FileStorageTest(TransactionTestCase):
    def setUp(self):
        self.file_storage: FileStorage = FileStorage.objects.create()
        super().setUp()

    def test_default_values(self):
        self.assertLessEqual(
            timezone.now() - self.file_storage.created, timedelta(seconds=1)
        )

    def test_has_storage_location(self):
        self.assertFalse(self.file_storage.has_storage_location("local"))
        storage_local = StorageLocation.objects.create(
            file_storage=self.file_storage, url="url", connector_name="local"
        )
        self.assertFalse(self.file_storage.has_storage_location("local"))
        storage_local.status = StorageLocation.STATUS_DONE
        storage_local.save()
        self.assertTrue(self.file_storage.has_storage_location("local"))

    def test_default_storage_location(self):
        self.assertIsNone(self.file_storage.default_storage_location)
        storage_s3: StorageLocation = StorageLocation.objects.create(
            file_storage=self.file_storage, url="url", connector_name="S3"
        )
        self.assertEqual(self.file_storage.default_storage_location, storage_s3)
        storage_s3.status = StorageLocation.STATUS_DONE
        storage_s3.save()
        self.assertEqual(self.file_storage.default_storage_location, storage_s3)
        storage_gcs: StorageLocation = StorageLocation.objects.create(
            file_storage=self.file_storage, url="url", connector_name="GCS"
        )
        self.assertEqual(self.file_storage.default_storage_location, storage_s3)
        storage_gcs.status = StorageLocation.STATUS_DONE
        storage_gcs.save()
        # GCS should be considered before S3 (default priority 100)
        self.assertEqual(self.file_storage.default_storage_location, storage_gcs)
        storage_local: StorageLocation = StorageLocation.objects.create(
            file_storage=self.file_storage, url="url", connector_name="local"
        )
        self.assertEqual(self.file_storage.default_storage_location, storage_gcs)
        storage_local.status = StorageLocation.STATUS_DONE
        storage_local.save()
        # Local location should be default (lowest priority)
        self.assertEqual(self.file_storage.default_storage_location, storage_local)

    def test_default_storage_location_nonexisting(self):
        StorageLocation.objects.create(
            file_storage=self.file_storage,
            url="url",
            connector_name="S3",
            status=StorageLocation.STATUS_DONE,
        )
        # This one should have default priority of 100.
        nonexisting = StorageLocation.objects.create(
            file_storage=self.file_storage,
            url="url",
            connector_name="nonexisting",
            status=StorageLocation.STATUS_DONE,
        )
        self.assertEqual(self.file_storage.default_storage_location, nonexisting)

    def test_get_path(self):
        with self.assertRaises(AttributeError):
            self.file_storage.get_path()
        StorageLocation.objects.create(
            file_storage=self.file_storage,
            url="url",
            connector_name="local",
            status=StorageLocation.STATUS_DONE,
        )
        StorageLocation.objects.create(
            file_storage=self.file_storage,
            url="url",
            connector_name="S3",
            status=StorageLocation.STATUS_DONE,
        )
        self.assertEqual(self.file_storage.get_path(), "/url")
        self.assertEqual(self.file_storage.get_path(filename="test"), "/url/test")
        self.assertEqual(self.file_storage.get_path(prefix="/prefix"), "/prefix/url")
        self.assertEqual(
            self.file_storage.get_path(prefix="/prefix", filename="test"),
            "/prefix/url/test",
        )

    def test_subpath(self):
        with self.assertRaises(AttributeError):
            self.assertEqual(self.file_storage.subpath, "url")
        StorageLocation.objects.create(
            file_storage=self.file_storage,
            url="s3url",
            connector_name="S3",
            status=StorageLocation.STATUS_DONE,
        )
        self.assertEqual(self.file_storage.subpath, "s3url")
        StorageLocation.objects.create(
            file_storage=self.file_storage,
            url="localurl",
            connector_name="local",
            status=StorageLocation.STATUS_DONE,
        )
        self.assertEqual(self.file_storage.subpath, "localurl")


class StorageLocationTest(TransactionTestCase):
    def setUp(self):
        self.file_storage: FileStorage = FileStorage.objects.create()
        super().setUp()

    def test_default_values(self):
        location: StorageLocation = StorageLocation.objects.create(
            file_storage=self.file_storage, url="url"
        )
        self.assertEqual(location.status, StorageLocation.STATUS_PREPARING)
        self.assertLessEqual(
            timezone.now() - location.last_update, timedelta(seconds=1)
        )

    def test_last_update(self):
        location: StorageLocation = StorageLocation.objects.create(
            file_storage=self.file_storage, url="url"
        )
        last_update = location.last_update
        location.status = StorageLocation.STATUS_DONE
        location.save()
        self.assertNotEqual(location.last_update, last_update)

    def test_default_object_manager(self):
        location_preparing: StorageLocation = StorageLocation.objects.create(
            file_storage=self.file_storage, url="preparing"
        )
        location_done: StorageLocation = StorageLocation.objects.create(
            file_storage=self.file_storage,
            url="done",
            status=StorageLocation.STATUS_DONE,
        )
        done_objects = StorageLocation.objects.all()
        all_objects = StorageLocation.all_objects.all()
        self.assertEqual(done_objects.count(), 1)
        self.assertIn(location_done, done_objects)
        self.assertEqual(all_objects.count(), 2)
        self.assertIn(location_preparing, all_objects)
        self.assertIn(location_done, all_objects)

    def test_unique_together(self):
        StorageLocation.objects.create(file_storage=self.file_storage, url="url")
        StorageLocation.objects.create(file_storage=self.file_storage, url="url1")
        StorageLocation.objects.create(
            file_storage=self.file_storage, url="url", connector_name="S3"
        )
        with self.assertRaises(IntegrityError):
            StorageLocation.objects.create(file_storage=self.file_storage, url="url")
        with self.assertRaises(IntegrityError):
            StorageLocation.objects.create(
                file_storage=self.file_storage, url="url", connector_name="S3"
            )

    def test_delete_data(self):
        path1 = ReferencedPath.objects.create(path="remove_me.txt")
        path2 = ReferencedPath.objects.create(path="dir/remove_me.txt")
        storage_location = StorageLocation.objects.create(
            file_storage=self.file_storage, url="url", connector_name="local"
        )
        storage_location.files.add(path1, path2)
        delete = MagicMock()
        connector_mock = MagicMock(delete=delete)
        with patch("resolwe.storage.models.connectors", {"local": connector_mock}):
            storage_location.delete_data()
        self.assertEqual(delete.call_count, 1)
        self.assertCountEqual(
            delete.call_args_list[0][0][1],
            ["remove_me.txt", "dir/remove_me.txt"],
        )

    def test_verify_data_ok(self):
        path1 = ReferencedPath.objects.create(
            path="1", md5="1", crc32c="1", awss3etag="1"
        )
        storage_location: StorageLocation = StorageLocation.objects.create(
            file_storage=self.file_storage, url="url", connector_name="local"
        )
        storage_location.files.add(path1)
        with patch(
            "resolwe.storage.models.StorageLocation.connector"
        ) as connector_mock:
            connector_mock.duplicate = MagicMock(
                return_value=MagicMock(check_url=MagicMock(return_value=True))
            )
            connector_mock.get_hashes = MagicMock(
                side_effect=[
                    {"md5": "1", "crc32c": "1", "awss3etag": "1"},
                ]
            )
            verified = storage_location.verify_data()
            self.assertTrue(verified)

    def test_verify_data_connector_failed(self):
        path1 = ReferencedPath.objects.create(
            path="1", md5="1", crc32c="1", awss3etag="1"
        )
        storage_location: StorageLocation = StorageLocation.objects.create(
            file_storage=self.file_storage, url="url", connector_name="local"
        )
        storage_location.files.add(path1)
        with patch(
            "resolwe.storage.models.StorageLocation.connector"
        ) as connector_mock:
            connector_mock.duplicate = MagicMock(
                return_value=MagicMock(check_url=MagicMock(return_value=False))
            )
            connector_mock.get_hashes = MagicMock(
                side_effect=[
                    {"md5": "1", "crc32c": "1", "awss3etag": "1"},
                ]
            )
            verified = storage_location.verify_data()
            self.assertFalse(verified)

    def test_verify_data_hash_failed(self):
        path1 = ReferencedPath.objects.create(
            path="1", md5="1", crc32c="1", awss3etag="1"
        )
        storage_location: StorageLocation = StorageLocation.objects.create(
            file_storage=self.file_storage, url="url", connector_name="local"
        )
        storage_location.files.add(path1)
        with patch(
            "resolwe.storage.models.StorageLocation.connector"
        ) as connector_mock:
            connector_mock.duplicate = MagicMock(
                return_value=MagicMock(check_url=MagicMock(return_value=False))
            )
            connector_mock.get_hashes = MagicMock(
                side_effect=[
                    {"md5": "1", "crc32c": "invalid", "awss3etag": "1"},
                ]
            )
            verified = storage_location.verify_data()
            self.assertFalse(verified)


class AccessLogTest(TransactionTestCase):
    def setUp(self):
        self.file_storage: FileStorage = FileStorage.objects.create()
        self.storage_location: StorageLocation = StorageLocation.objects.create(
            file_storage=self.file_storage, url="url"
        )
        super().setUp()

    def test_default_values(self):
        log: AccessLog = AccessLog.objects.create(
            storage_location=self.storage_location, reason="reason"
        )
        self.assertIsNone(log.finished)
        self.assertLessEqual(timezone.now() - log.started, timedelta(seconds=1))



""".. Ignore pydocstyle D400.

========
Listener
========

Standalone client used as a contact point for executors. It uses zeromq
socket for communication.

.. autoclass:: resolwe.flow.managers.listener.ExecutorListener
    :members:

"""

import asyncio
import copy
import json
import logging
import os
import shlex
import threading
import time
from collections import defaultdict
from contextlib import suppress
from importlib import import_module
from pathlib import Path
from typing import Any, Dict, List, MutableMapping, Optional, Set, Tuple, Union

import zmq
import zmq.asyncio
from asgiref.sync import async_to_sync
from channels.db import database_sync_to_async

from django.conf import settings
from django.core.exceptions import PermissionDenied, ValidationError
from django.forms.models import model_to_dict
from django.utils.timezone import now

from django_priority_batch import PrioritizedBatcher

from resolwe.flow.engine import BaseEngine, InvalidEngineError, load_engines
from resolwe.flow.executors import constants
from resolwe.flow.executors.socket_utils import (
    BaseProtocol,
    Message,
    PeerIdentity,
    PeerStatus,
    Response,
    ResponseStatus,
)
from resolwe.flow.executors.zeromq_utils import ZMQCommunicator
from resolwe.flow.managers import consumer
from resolwe.flow.managers.protocol import ExecutorFiles, WorkerProtocol
from resolwe.flow.models import Data, Process, Storage, Worker
from resolwe.flow.utils import iterate_schema
from resolwe.storage import settings as storage_settings
from resolwe.storage.models import AccessLog
from resolwe.test.utils import is_testing
from resolwe.utils import BraceMessage as __

# Register plugins by importing them.
from .basic_commands_plugin import BasicCommands  # noqa: F401
from .init_container_plugin import InitContainerPlugin  # noqa: F401
from .plugin import plugin_manager
from .python_process_plugin import PythonProcess  # noqa: F401

logger = logging.getLogger(__name__)


class Processor:
    """Class respresents instance of one connected worker."""

    def __init__(
        self,
        peer_identity: PeerIdentity,
        starting_sequence_number: int,
        listener: "ListenerProtocol",
    ):
        """Initialize.

        Assumption: first_message is checked to be of the expected type.

        :raises ValueError: when identity is not in the correct format.
        :raises Data.DoesNotExist: when Data object is not found.
        """
        self.peer_identity = peer_identity
        self.data_id = int(peer_identity)
        self._data: Optional[Data] = None
        self._worker: Optional[Worker] = None
        self._storage_fields: Optional[Set[str]] = None
        self._listener = listener
        self.expected_sequence_number = starting_sequence_number
        self.return_code = 0

    @property
    def contributor(self) -> settings.AUTH_USER_MODEL:
        """Get the user that created the Data object we are processing.

        This user is used when checking permissions.
        """
        return self.data.contributor

    @property
    def data(self) -> Data:
        """Get the data object.

        The evaluation is done when needed in order not to make potentially
        failing SQL query in the constructor.

        This method uses Django ORM.
        """
        if self._data is None:
            self._data = Data.objects.get(pk=self.data_id)
        return self._data

    @property
    def worker(self) -> Worker:
        """Get the worker object.

        The evaluation is done when needed in order not to make potentially
        failing SQL query in the constructor.

        This method uses Django ORM.
        """
        if self._worker is None:
            self._worker = Worker.objects.get(data=self.data)
        return self._worker

    @property
    def storage_fields(self) -> Set[str]:
        """Get the names of storage fields in schema."""
        if self._storage_fields is None:
            self._storage_fields = {
                field_name
                for schema, _, field_name in iterate_schema(
                    self.data.output, self.data.process.output_schema
                )
                if schema["type"].startswith("basic:json:")
            }
        return self._storage_fields

    def save_storage(self, key: str, content: str, data: Optional[Data] = None):
        """Save storage to data object.

        It is the responsibility of the caller to save the changes to the Data
        object.
        """
        data = data or self.data
        if key in data.output:
            storage = Storage.objects.get(pk=data.output[key])
            storage.json = content
            storage.save()
        else:
            storage = Storage.objects.create(
                json=content,
                name="Storage for data id {}".format(data.pk),
                contributor=data.contributor,
            )
            storage.data.add(data)
            data.output[key] = storage.pk

    def _unlock_all_inputs(self):
        """Unlock all data objects that were locked by the given data.

        If exception occurs during unlocking we can not do much but ignore it.
        The method is called when dispatcher is nudged.
        """
        with suppress(Exception):
            query = AccessLog.objects.filter(cause_id=self.data_id)
            query.update(finished=now())

    def _choose_worst_status(self, status1: str, status2: str) -> str:
        """Get more problematic status of the two.

        Assumption: statuses in Data.STATUS_CHOICES are ordered from least
            to most problematic.

        :raises AssertionError: when status1 or status2 do not represent
            possible status of Data object.

        """
        statuses = [status[0] for status in Data.STATUS_CHOICES]
        assert status1 in statuses
        assert status2 in statuses
        return max([status1, status2], key=lambda status: statuses.index(status))

    def _save_error(self, error: str):
        """Log error to Data object, ignore possible exceptions.

        The status of Data object is changed to ``Data.STATUS_ERROR``.
        """
        with suppress(Exception):
            self.data.refresh_from_db()
        self.data.process_error.append(error)
        self.data.status = Data.STATUS_ERROR
        with suppress(Exception):
            self.data.save(update_fields=["process_error", "status"])

    def _log_exception(
        self, error: str, extra: Optional[dict] = None, save_to_data_object: bool = True
    ):
        """Log current exception and optionally store it to the data object.

        When argument ``save_to_data_object`` is false the status of the
        data object is not changed.
        """
        if extra is None:
            extra = dict()
        extra.update({"data_id": self.data.id})
        logger.exception(error, extra=extra)
        if save_to_data_object:
            self._save_error(error)

    def _log_error(
        self, error: str, extra: dict = {}, save_to_data_object: bool = True
    ):
        """Log error and optionally store it to the data object.

        When argument ``save_to_data_object`` is false the status of the
        data object is not changed.
        """
        extra.update({"data_id": self.data.id})
        logger.error(error, extra=extra)
        if save_to_data_object:
            self._save_error(error)

    def _abort_processing(self):
        """Abort processing.

        Use this call with caution. It also notifier the dispatcher to abort
        processing which means closing listener connection while testing.

        A sensible default is to call this method only when peer becomes
        unresponsive.
        """
        self.notify_dispatcher_abort()
        self._listener.remove_peer(str(self.data_id).encode())

    async def _finish_processing(self):
        """Finish processing.

        This method notifies the dispatcher that the worker has finished
        processing the data object.
        """
        self.notify_dispatcher_finish()

    def _update_data(self, changes: Dict[str, Any], refresh: bool = False):
        """Update the data object.

        Changes are dictionary mapping from field names to field values.

        :param changes: dictionary with chenges to be saved.
        :param refresh: if True refresh data object before applying changes.

        :raises: exception when data object cannot be saved.
        """
        if "process_rc" in changes and changes["process_rc"] < 0:
            changes["process_rc"] = abs(changes["process_rc"])
        self._update_object(changes, self.data)

    def _update_worker(self, changes: Dict[str, Any]):
        """Update the worker object.

        :raises: exception when data object cannot be saved.
        """
        self._update_object(changes, self.worker)

    def _update_object(self, changes: Dict[str, Any], obj: Union[Worker, Data]):
        """Update object properties and save them to database."""
        for key, value in changes.items():
            setattr(obj, key, value)
        obj.save(update_fields=list(changes.keys()))

    async def terminate(self):
        """Send the terminate command to the worker.

        Peer should terminate by itself and send finish message back to us.
        """
        await database_sync_to_async(self._save_error, thread_sensitive=False)(
            "Processing was cancelled."
        )
        # Ignore the possible timeout.
        with suppress(RuntimeError):
            await self._listener.communicator.send_command(
                Message.command("terminate", ""),
                peer_identity=str(self.data_id).encode(),
                response_timeout=1,
            )

    async def peer_not_responding(self):
        """Peer is not responding, abort the processing.

        TODO: what to do with unresponsive peer which wakes up later and
        starts sending commands to the listener? Currently the situation
        is handled by sending "terminate" command to it and ignore its
        commands.
        """
        logger.debug(__("Peer with id={} is not responding.", self.data_id))
        await database_sync_to_async(self._log_error)("Worker is not responding")
        await database_sync_to_async(self._update_worker)(
            {"status": Worker.STATUS_NONRESPONDING}
        )
        await self.notify_dispatcher_abort_async()
        self._listener.remove_peer(str(self.data_id).encode())

    def process_command(self, message: Message) -> Response:
        """Process a single command from the peer.

        This command is run in the database_sync_to_async so it is safe to
        perform Django ORM operations inside.

        Exceptions will be handler one level up and error response will be
        sent in this case.
        """
        # This worker must be in status processing or preparing.
        # All messages from workers not in this status will be discarted and
        # error will be returned.
        if self.worker.status not in [
            Worker.STATUS_PROCESSING,
            Worker.STATUS_PREPARING,
        ]:
            self._log_error(
                f"Wrong worker status: {self.worker.status} for peer with id {self.data_id}."
            )
            return message.respond_error(f"Wrong worker status: {self.worker.status}")

        command_name = message.command_name
        handler_name = f"handle_{command_name}"
        handler = plugin_manager.get_handler(command_name)
        if not handler:
            error = f"No command handler for '{command_name}'."
            self._log_error(error, save_to_data_object=False)
            return message.respond_error(error)

        # Read sequence number and refresh data object if it differs.
        if self.expected_sequence_number != message.sequence_number:
            try:
                self.data.refresh_from_db()
                self.worker.refresh_from_db()
            except:
                self._log_exception("Unable to refresh data object")
                return message.respond_error("Unable to refresh the data object")

        if self.worker.status != Worker.STATUS_PROCESSING:
            self.worker.status = Worker.STATUS_PROCESSING
            self.worker.save(update_fields=["status"])

        if self.data.started is None:
            self.data.started = now()
            self.data.save(update_fields=["started"])

        self.expected_sequence_number = message.sequence_number + 1
        try:
            with PrioritizedBatcher.global_instance():
                result = handler(message, self)
                # Set status of the response to ERROR when data object status
                # is Data.STATUS_ERROR. Such response will trigger terminate
                # procedure in the processing container and stop processing.
                if self.data.status == Data.STATUS_ERROR:
                    result.type_data = ResponseStatus.ERROR.value
                return result
        except ValidationError as err:
            error = (
                f"Validation error when saving Data object of process "
                f"'{self.data.process.slug}' ({handler_name}): "
                f"{err}"
            )
            self._log_exception(error)
            return message.respond_error("Validation error")
        except Exception as err:
            error = f"Error in command handler '{handler_name}': {err}"
            self._log_exception(error)
            return message.respond_error(f"Error in command handler '{handler_name}'")

    def notify_dispatcher_abort(self):
        """Notify dispatcher that processing was aborted.

        .. IMPORTANT::

            This only makes manager's state consistent and doesn't
            affect Data object in any way.

        :param obj: The Channels message object. Command object format:

            .. code-block:: none

                {
                    'command': 'abort',
                    'data_id': [id of the :class:`~resolwe.flow.models.Data` object
                               this command was triggered by],
                }
        """
        async_to_sync(self.notify_dispatcher_abort_async)()

    async def notify_dispatcher_abort_async(self):
        """Notify dispatcher that processing was aborted.

        :param obj: The Channels message object. Command object format:

            .. code-block:: none

                {
                    'command': 'abort',
                    'data_id': [id of the :class:`~resolwe.flow.models.Data` object
                               this command was triggered by],
                }
        """
        await database_sync_to_async(self._unlock_all_inputs)()
        await consumer.send_event(
            {
                WorkerProtocol.COMMAND: WorkerProtocol.ABORT,
                WorkerProtocol.DATA_ID: self.data.id,
                WorkerProtocol.FINISH_COMMUNICATE_EXTRA: {
                    "executor": getattr(settings, "FLOW_EXECUTOR", {}).get(
                        "NAME", "resolwe.flow.executors.local"
                    ),
                },
            }
        )
        logger.debug("notify_dispatcher_abort: consumer event sent")

    def notify_dispatcher_finish(self):
        """Notify dispatcher that the processing is finished.

        See ``notify_dispatcher_abort`` for message format.
        """
        async_to_sync(self.notify_dispatcher_finish_async)()

    async def notify_dispatcher_finish_async(self):
        """Notify dispatcher that the processing is finished.

        See ``notify_dispatcher_abort`` for message format.
        """
        await database_sync_to_async(self._unlock_all_inputs)()
        await consumer.send_event(
            {
                WorkerProtocol.COMMAND: WorkerProtocol.FINISH,
                WorkerProtocol.DATA_ID: self.data.id,
                WorkerProtocol.FINISH_COMMUNICATE_EXTRA: {
                    "executor": getattr(settings, "FLOW_EXECUTOR", {}).get(
                        "NAME", "resolwe.flow.executors.local"
                    ),
                },
            }
        )


class ListenerProtocol(BaseProtocol):
    """Listener protocol implementation."""

    def __init__(
        self,
        hosts: List[str],
        port: int,
        protocol: str,
        zmq_socket: Optional[zmq.asyncio.Socket] = None,
    ):
        """Initialize."""
        if zmq_socket is None:
            zmq_context: zmq.asyncio.Context = zmq.asyncio.Context.instance()
            zmq_socket = zmq_context.socket(zmq.ROUTER)
            zmq_socket.setsockopt(zmq.ROUTER_HANDOVER, 1)
            for host in hosts:
                zmq_socket.bind(f"{protocol}://{host}:{port}")

        super().__init__(
            ZMQCommunicator(
                zmq_socket, "listener <-> workers", logger, self.peer_status_changed
            ),
            logger,
        )

        # Mapping from the data ID to the Worker instance. Each data object is
        # processed by (at most) one Worker so the map is one-to-one.
        self.peers: MutableMapping[PeerIdentity, Processor] = dict()

        # The code above is used when preparing script for the worker when it
        # is requested. Maybe there is a better way?
        self._execution_engines = self._load_execution_engines()
        logger.info(
            __(
                "Found {} execution engines: {}",
                len(self._execution_engines),
                ", ".join(self._execution_engines.keys()),
            )
        )
        self._expression_engines = self._load_expression_engines()
        logger.info(
            __(
                "Found {} expression engines: {}",
                len(self._expression_engines),
                ", ".join(self._expression_engines.keys()),
            )
        )
        self._executor_preparer = self._load_executor_preparer()

        self._parallel_commands_counter = 0
        self._max_parallel_commands = 10
        self._parallel_commands_semaphore: Optional[asyncio.Semaphore] = None
        self._get_program_lock = threading.Lock()

        self._bootstrap_cache: Dict[str, Any] = defaultdict(dict)

    def handle_log(
        self, message: Message[Union[str, bytes, bytearray]]
    ) -> Response[str]:
        """Handle an incoming log processing request.

        :param obj: The Channels message object. Command object format:

            .. code-block:: none

                {
                    'command': 'log',
                    'message': [log message]
                }
        """
        record_dict = json.loads(message.message_data)
        record_dict["args"] = tuple(record_dict["args"])

        executors_dir = Path(__file__).parents[1] / "executors"
        record_dict["pathname"] = os.fspath(executors_dir / record_dict["pathname"])
        logger.handle(logging.makeLogRecord(record_dict))
        return message.respond_ok("OK")

    def _bootstrap_prepare_static_cache(self):
        """Prepare cache for bootstrap."""

        def marshal_settings() -> dict:
            """Marshal Django settings into a serializable object.

            :return: The serialized settings.
            :rtype: dict
            """
            result = {}
            for key in dir(settings):
                if any(
                    map(
                        key.startswith,
                        [
                            "FLOW_",
                            "RESOLWE_",
                            "CELERY_",
                            "KUBERNETES_",
                        ],
                    )
                ):
                    result[key] = getattr(settings, key)
            result.update(
                {
                    "USE_TZ": settings.USE_TZ,
                    "FLOW_EXECUTOR_TOOLS_PATHS": self._executor_preparer.get_tools_paths(),
                    "FLOW_STORAGE": storage_settings.FLOW_STORAGE,
                }
            )
            # TODO: this is q&d solution for serializing Path objects.
            return json.loads(json.dumps(result, default=str))

        # Prepare Django settings.
        if "settings" not in self._bootstrap_cache:
            self._bootstrap_cache["settings"] = marshal_settings()
            connectors_settings = copy.deepcopy(storage_settings.STORAGE_CONNECTORS)
            for connector_settings in connectors_settings.values():
                # Fix class name for inclusion in the executor.
                klass = connector_settings["connector"]
                klass = "executors." + klass.rsplit(".storage.")[-1]
                connector_settings["connector"] = klass
                connector_config = connector_settings["config"]
                # Prepare credentials for executor.
                if "credentials" in connector_config:
                    src_credentials = connector_config["credentials"]
                    base_credentials_name = os.path.basename(src_credentials)

                    self._bootstrap_cache["connector_secrets"][
                        base_credentials_name
                    ] = ""
                    if os.path.isfile(src_credentials):
                        with open(src_credentials, "r") as f:
                            self._bootstrap_cache["connector_secrets"][
                                base_credentials_name
                            ] = f.read()
                    connector_config["credentials"] = os.fspath(
                        constants.SECRETS_VOLUME / base_credentials_name
                    )
            self._bootstrap_cache["settings"][
                "STORAGE_CONNECTORS"
            ] = connectors_settings
            self._bootstrap_cache["settings"][
                "FLOW_VOLUMES"
            ] = storage_settings.FLOW_VOLUMES

            # Prepare process meta data.
            self._bootstrap_cache["process_meta"] = {
                k: getattr(Process, k)
                for k in dir(Process)
                if k.startswith("SCHEDULING_CLASS_")
                and isinstance(getattr(Process, k), str)
            }

            self._bootstrap_cache["process"] = dict()

    def bootstrap_prepare_process_cache(self, data: Data):
        """Prepare cache for process with the given id."""
        if data.process_id not in self._bootstrap_cache["process"]:
            self._bootstrap_cache["process"][data.process_id] = model_to_dict(
                data.process
            )
            self._bootstrap_cache["process"][data.process_id][
                "resource_limits"
            ] = data.process.get_resource_limits()

    def handle_bootstrap(self, message: Message[Tuple[int, str]]) -> Response[Dict]:
        """Handle bootstrap request.

        :raises RuntimeError: when settings name is not known.
        """
        data_id, settings_name = message.message_data
        data = Data.objects.get(pk=data_id)
        if is_testing():
            self._bootstrap_cache = defaultdict(dict)
        self._bootstrap_prepare_static_cache()
        response: Dict[str, Any] = dict()
        self.bootstrap_prepare_process_cache(data)

        if settings_name == "executor":
            execution_engine_name = data.process.run.get("language", None)
            execution_engine = self._get_execution_engine(execution_engine_name)
            volume_maps = execution_engine.prepare_volumes()

            response[ExecutorFiles.EXECUTOR_SETTINGS] = {
                "DATA_DIR": data.location.get_path()
            }
            response[ExecutorFiles.LOCATION_SUBPATH] = data.location.subpath
            response[ExecutorFiles.DJANGO_SETTINGS] = self._bootstrap_cache["settings"]
            response[ExecutorFiles.DJANGO_SETTINGS]["RUNTIME_VOLUME_MAPS"] = volume_maps
            response[ExecutorFiles.PROCESS_META] = self._bootstrap_cache["process_meta"]
            response[ExecutorFiles.PROCESS] = self._bootstrap_cache["process"][
                data.process.id
            ]

        elif settings_name == "init":
            response[ExecutorFiles.DJANGO_SETTINGS] = {
                "STORAGE_CONNECTORS": self._bootstrap_cache["settings"][
                    "STORAGE_CONNECTORS"
                ],
                "FLOW_STORAGE": storage_settings.FLOW_STORAGE,
            }
            if hasattr(settings, constants.INPUTS_VOLUME_NAME):
                response[ExecutorFiles.DJANGO_SETTINGS][
                    constants.INPUTS_VOLUME_NAME
                ] = self._bootstrap_cache["settings"][constants.INPUTS_VOLUME_NAME]
            response[ExecutorFiles.SECRETS_DIR] = self._bootstrap_cache[
                "connector_secrets"
            ]
            try:
                response[ExecutorFiles.SECRETS_DIR].update(data.resolve_secrets())
            except PermissionDenied as e:
                data.process_error.append(str(e))
                data.save(update_fields=["process_error"])
                raise
            response[ExecutorFiles.LOCATION_SUBPATH] = data.location.subpath
        elif settings_name == "communication":
            response[ExecutorFiles.DJANGO_SETTINGS] = {
                "STORAGE_CONNECTORS": self._bootstrap_cache["settings"][
                    "STORAGE_CONNECTORS"
                ],
                "FLOW_STORAGE": storage_settings.FLOW_STORAGE,
            }
            response[ExecutorFiles.LOCATION_SUBPATH] = data.location.subpath
        else:
            raise RuntimeError(
                f"Settings {settings_name} sent by peer with id {data_id} unknown."
            )
        return message.respond_ok(response)

    async def peer_status_changed(
        self, peer_identity: PeerIdentity, status: PeerStatus
    ):
        """Check the new peer status and invoke the necessary handlers.

        Remove the unresponsive peer from the list of peers and update its
        status to UNRESPONSIVE.
        """
        logger.debug(__("Peer {} status changed: {}.", peer_identity, status))
        if status == PeerStatus.UNRESPONSIVE:
            peer = self.remove_peer(peer_identity)
            if peer is not None:
                await peer.peer_not_responding()

    def remove_peer(self, peer_identity: PeerIdentity) -> Optional[Processor]:
        """Remove peer from list of known peers."""
        peer = self.peers.pop(peer_identity, None)
        if peer is None:
            logger.error("Peer %s is not known, can not remove.", peer_identity)
        return peer

    async def _get_response(
        self, peer_identity: PeerIdentity, message: Message
    ) -> Response:
        """Handle command received over 0MQ."""
        # Logging and bootstraping are handled separately.
        if message.command_name in ["log", "bootstrap"]:
            try:
                handler = getattr(self, f"handle_{message.command_name}")
                return await database_sync_to_async(handler, thread_sensitive=False)(
                    message
                )
            except Exception:
                logger.exception(
                    __(
                        "Error handling command {} for peer {}.",
                        message.command_name,
                        peer_identity,
                    )
                )
                return message.respond("Error in handler.", ResponseStatus.ERROR)

        if peer_identity not in self.peers:
            try:
                peer = Processor(
                    peer_identity,
                    message.sequence_number,
                    self,
                )
                self.peers[peer_identity] = peer
            except:
                logger.exception("Error creating Processor instance.")
                return message.respond(
                    "Error creating Processor instance", ResponseStatus.ERROR
                )

        peer = self.peers[peer_identity]
        try:
            response = await database_sync_to_async(
                peer.process_command, thread_sensitive=False
            )(message)
        except:
            logger.exception("Error in process_command method.")
            return message.respond("Error processing command", ResponseStatus.ERROR)
        else:
            return response

    async def _process_command(
        self, peer_identity: PeerIdentity, received_message: Message
    ):
        """Process command."""
        # Executor uses separate identity of the form f"e_{data.id}".
        response_identity = peer_identity
        if b"_" in peer_identity:
            peer_identity = peer_identity.split(b"_")[1]

        if self._parallel_commands_semaphore is None:
            self._parallel_commands_semaphore = asyncio.Semaphore(
                self._max_parallel_commands
            )

        self.logger.debug(
            __("Internal semaphore count: {}", self._parallel_commands_semaphore._value)
        )
        async with self._parallel_commands_semaphore:
            self._parallel_commands_counter += 1
            self.logger.debug(
                __("Processing {} commands", self._parallel_commands_counter)
            )
            response = await self._get_response(peer_identity, received_message)
            try:
                assert received_message.sent_timestamp is not None
                response_time = time.time() - received_message.sent_timestamp
                self.logger.debug(__("Response time: {}", response_time))
                await self.communicator.send_response(response, response_identity)

            except RuntimeError:
                # When unable to send message just log the error. Peer will be
                # removed by the watchdog later on.
                self.logger.exception(
                    __("Protocol: error sending response to {}.", received_message)
                )
            try:
                # Remove peer and nudge manager on "finish" command.
                if received_message.command_name == "finish":
                    peer = self.peers[peer_identity]
                    self.remove_peer(peer_identity)

                    # Worker status must be updated in the dispatcher to avoid
                    # synchronization issues while testing (if status is updated here
                    # another message might raise runtime barrier before our message
                    # reaches the dispatcher.
                    with suppress(Data.DoesNotExist):
                        await peer.notify_dispatcher_finish_async()
            except:
                logger.exception(
                    __("Error processing 'finish' for peer {}", peer_identity)
                )
            finally:
                self._parallel_commands_counter -= 1
                logger.debug(
                    __(
                        "Finished processing command, {} remaining.",
                        self._parallel_commands_counter,
                    )
                )

    async def process_command(
        self, peer_identity: PeerIdentity, received_message: Message
    ):
        """Override process command in socket_utils."""
        # Make it run in the background so another message can be processed.
        asyncio.ensure_future(self._process_command(peer_identity, received_message))

    def get_program(self, data: Data) -> str:
        """Get a program for given data object."""
        # When multiple get_script commands run in parallel the string
        # escaping will be fragile at best. Process this command here
        # without offloading it to the listener.
        with self._get_program_lock:
            execution_engine_name = data.process.run.get("language", None)
            program = [self._get_execution_engine(execution_engine_name).evaluate(data)]
            # TODO: should executor be changed? Definitely for tests (from case to case).
            # Should I use file prepared by the Dispatcher (that takes care of that)?
            env_vars = self.get_executor().get_environment_variables()
            settings_env_vars = {
                "RESOLWE_HOST_URL": getattr(settings, "RESOLWE_HOST_URL", "localhost"),
            }
            additional_env_vars = getattr(settings, "FLOW_EXECUTOR", {}).get(
                "SET_ENV", {}
            )
            env_vars.update(settings_env_vars)
            env_vars.update(additional_env_vars)
            export_commands = [
                "export {}={}".format(key, shlex.quote(value))
                for key, value in env_vars.items()
            ]

            # Paths for Resolwe tools.
            tools_num = len(self._executor_preparer.get_tools_paths())
            tools_paths = ":".join(
                f"/usr/local/bin/resolwe/{i}" for i in range(tools_num)
            )
            export_commands += [f"export PATH=$PATH:{tools_paths}"]

            # Disable brace expansion and set echo.
            echo_commands = ["set -x +B"]

            commands = echo_commands + export_commands + program
            return os.linesep.join(commands)

    def _get_execution_engine(self, name):
        """Return an execution engine instance."""
        try:
            return self._execution_engines[name]
        except KeyError:
            raise InvalidEngineError("Unsupported execution engine: {}".format(name))

    def _load_execution_engines(self) -> Dict[str, BaseEngine]:
        """Load execution engines."""
        execution_engines = getattr(
            settings, "FLOW_EXECUTION_ENGINES", ["resolwe.flow.execution_engines.bash"]
        )
        return load_engines(
            self, "ExecutionEngine", "execution_engines", execution_engines
        )

    def _load_expression_engines(self) -> Dict[str, BaseEngine]:
        """Load expression engines."""
        expression_engines = getattr(
            settings,
            "FLOW_EXPRESSION_ENGINES",
            ["resolwe.flow.expression_engines.jinja"],
        )
        return load_engines(
            self, "ExpressionEngine", "expression_engines", expression_engines
        )

    def _load_executor_preparer(self) -> Any:
        """Load and return the executor preparer class."""
        executor_name = (
            getattr(settings, "FLOW_EXECUTOR", {}).get(
                "NAME", "resolwe.flow.executors.docker"
            )
            + ".prepare"
        )
        return import_module(executor_name).FlowExecutorPreparer()

    def get_executor(self) -> Any:
        """Get the executor preparer class."""
        return self._executor_preparer

    def get_expression_engine(self, name: str):
        """Return an expression engine instance."""
        try:
            return self._expression_engines[name]
        except KeyError:
            raise InvalidEngineError("Unsupported expression engine: {}".format(name))


def handle_exceptions(loop, context):
    """Log uncaught exceptions in asyncio."""
    msg = context.get("exception", context["message"])
    name = context.get("future").get_coro().__name__
    logger.error(f"Caught exception from {name}: {msg}")


class ExecutorListener:
    """The contact point implementation for executors."""

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        """Initialize."""
        # Read socket parameters.
        self._hosts = list(
            kwargs.get(
                "hosts",
                getattr(settings, "LISTENER_CONNECTION", {}).get(
                    "hosts", {"local": "127.0.0.1"}
                ),
            ).values()
        )

        self.port = kwargs.get(
            "port", getattr(settings, "LISTENER_CONNECTION", {}).get("port", 53892)
        )
        self.protocol = kwargs.get(
            "protocol",
            getattr(settings, "LISTENER_CONNECTION", {}).get("protocol", "tcp"),
        )

        # When zmq_socket kwarg is not None, use this one instead of creating
        # a new one.
        self.zmq_socket = kwargs.get("zmq_socket")

        self._listener_protocol: Optional[ListenerProtocol] = None

        # Running coordination.
        self._should_stop: Optional[asyncio.Event] = None
        self._runner_future: Optional[asyncio.Future] = None
        self._communicating_future: Optional[asyncio.Future] = None

    @property
    def should_stop(self):
        """Return should stop event.

        Use lazy instantiation to avoid creation event loop on init.
        """
        if self._should_stop is None:
            self._should_stop = asyncio.Event()
        return self._should_stop

    @property
    def hosts(self):
        """Get hosts to bind on."""
        return self._hosts

    @hosts.setter
    def hosts(self, hosts: dict):
        """Set the hosts to bind on."""
        self._hosts = list(hosts.values())

    @property
    def listener_protocol(self):
        """Return the listener protocol object.

        Used to lazy create object when property is accessed.
        """
        if self._listener_protocol is None:
            self._listener_protocol = ListenerProtocol(
                self.hosts, self.port, self.protocol, self.zmq_socket
            )
        return self._listener_protocol

    async def terminate_worker(self, identity: PeerIdentity):
        """Terminate the worker."""
        peer = self.listener_protocol.peers.get(identity)
        if peer is not None:
            await peer.terminate()

    async def __aenter__(self):
        """On entering a context, start the listener thread."""
        loop = asyncio.get_event_loop()
        loop.set_exception_handler(handle_exceptions)
        self._should_stop = None
        self._runner_future = asyncio.ensure_future(self.run())
        return self

    async def __aexit__(self, typ, value, trace):
        """On exiting a context, kill the listener and wait for it.

        .. note::

            Exceptions are all propagated.
        """
        assert self._runner_future is not None
        self.terminate()
        await asyncio.gather(self._runner_future)
        self._listener_protocol = None

    def terminate(self):
        """Stop the standalone manager."""
        logger.info(__("Terminating Resolwe listener."))
        self.should_stop.set()

    async def run(self):
        """Run the main listener run loop.

        Doesn't return until :meth:`terminate` is called.
        """
        logger.info(
            __(
                f"Starting Resolwe listener on  '{self.protocol}://{self.hosts}:{self.port}'."
            )
        )
        communicator_future = asyncio.ensure_future(
            self.listener_protocol.communicate()
        )
        await self.should_stop.wait()
        self.listener_protocol.stop_communicate()
        await communicator_future
        logger.info(
            __(
                f"Stoping Resolwe listener on  '{self.protocol}://{self.hosts}:{self.port}'."
            )
        )



"""Python process Django Models."""
import json
import os
from pathlib import Path
from typing import Any, Dict, List, MutableMapping, Optional, Type

from .communicator import communicator
from .fields import (
    BooleanField,
    DateTimeField,
    Field,
    IntegerField,
    JsonField,
    ListField,
    RelationDescriptor,
    StringField,
    fields_from_schema,
)

DATA_ID = int(os.getenv("DATA_ID", "-1"))


class ModelField(Field):
    """Field representing a model.

    ForeignKey relation is modeled with this field.
    """

    field_type = "model"

    def __init__(self, full_model_name: str, *args, **kwargs):
        """Initialize."""
        super().__init__(*args, **kwargs)
        self.full_model_name = full_model_name
        self.app_name, self.related_model_name = full_model_name.split(".")
        self.label = full_model_name
        self.known_models = RegisteredModels()

    def to_python(self, value):
        """Return the python object representing the field."""
        if isinstance(value, Model):
            return value
        if isinstance(value, int):
            model_class = self.known_models.get_model(self.full_model_name)
            return model_class(value)
        return super().to_python(value)

    def to_output(self, value):
        """Convert the value to output format."""
        if isinstance(value, int):
            return value
        if isinstance(value, Model):
            return value.id
        if self.related_model_name == "Storage" and isinstance(value, str):
            return value
        return super().to_output(value)


FIELDS_MAP = {
    "AutoField": IntegerField,
    "VersionField": StringField,
    "CharField": StringField,
    "TextField": StringField,
    "DateTimeField": DateTimeField,
    "ForeignKey": ModelField,
    "ManyToManyField": ListField,
    "PositiveIntegerField": IntegerField,
    "PositiveSmallIntegerField": IntegerField,
    "ArrayField": Field,
    "JSONField": JsonField,
    "BigIntegerField": IntegerField,
    "BooleanField": BooleanField,
    "SlugField": StringField,
}

_hydrate_cache: Dict[int, str] = dict()


def hydrate_if_needed(value, model_instance, field_name, field):
    """Hydrate path if needed."""
    # Only hidrate paths on outputs of data objects that we are not processing.
    if (
        model_instance._model_name == "Data"
        and model_instance.id != DATA_ID
        and field_name == "output"
        and field.get_field_type()
        in ("basic:file", "list:basic:file", "basic:dir", "list:basic:dir")
    ):
        if model_instance.id not in _hydrate_cache:
            _hydrate_cache[model_instance.id] = Path(
                communicator.resolve_data_path(model_instance.id)
            )
        base_path = _hydrate_cache[model_instance.id] / str(model_instance.location_id)
        if field.get_field_type().startswith("basic"):
            value.path = os.fspath(base_path / value.path)
        else:
            for entry in value:
                entry.path = os.fspath(base_path / entry.path)
    return value


class JSONDescriptor(MutableMapping[str, Any]):
    """JSON field."""

    def __init__(
        self,
        model: "Model",
        field_name: str,
        field_schema: List[Dict],
        read_only=False,
        cache: Optional[Dict[str, Any]] = None,
    ):
        """Initialize."""
        self._model = model
        self._model_name = self._model._model_name
        self._app_name = self._model._app_name
        self._field_name = field_name
        self._pk = self._model._pk
        self._cache: Dict[str, Any] = dict()
        self._read_only = read_only
        self._fields: Dict[str, Field] = dict()
        for field_name, field in fields_from_schema(field_schema).items():
            # JSON fields in schema are special since they represent the
            # storage model.
            if field.get_field_type() == "basic:json":
                field = ModelField(full_model_name="flow.Storage")
            field.contribute_to_class(self, self._fields, field_name)
        # Create cache if initial data is given.
        if cache is not None:
            self.refresh_cache(cache)

    def freeze(self):
        """Freeze fields."""
        self._read_only = True
        return self

    def refresh_cache(self, json_data: Optional[dict] = None):
        """Get the entire JSON field content from the server.

        TODO: read only required parts when Django supports it.
        """
        json_data = (
            json_data
            or communicator.get_model_fields(
                self._app_name, self._model_name, self._pk, [self._field_name]
            )[self._field_name]
        )
        for field_name, field in self._fields.items():
            if field_name in json_data:
                self._cache[field_name] = hydrate_if_needed(
                    field.clean(json_data[field_name]),
                    self._model,
                    self._field_name,
                    field,
                )

    def __getattr__(self, key):
        """Allow dot syntax."""
        return self.__getitem__(key)

    def __setattr__(self, key, value):
        """Set the value for the given key."""
        if key.startswith("_"):
            super().__setattr__(key, value)
        else:
            self.__setitem__(key, value)

    def __getitem__(self, key):
        """Get the value for the given key."""
        if self._fields is not None and key not in self._fields:
            raise AttributeError(f"Value {key} must be in the fields")
        if key not in self._cache:
            self.refresh_cache()
        # Return None even if key is not known.
        return self._cache.get(key)

    def __setitem__(self, key, value):
        """Set the value for the given key."""
        if self._read_only:
            raise ValueError("Cannot change read-only mapping.")
        if key not in self._fields:
            raise AttributeError(f"No field named {key}")
        to_output = None
        field = self._fields[key]
        if getattr(field, "related_model_name", None) == "Storage":
            # Check if JSON must be read from the file.
            if isinstance(value, str):
                json_file = Path(value)
                if json_file.is_file():
                    value = json.loads(json_file.read_text())
            storage = getattr(self, key)
            if storage is None:
                storage = Storage.create(
                    json=value,
                    name="Storage for data id {}".format(self._pk),
                    contributor=self._model.contributor,
                )
                storage.data += [self._model]
                value = storage
                to_output = storage.id
            else:
                storage.json = value
        else:
            value = field.clean(value)
            to_output = field.to_output(value)

        if to_output is not None:
            communicator.update_model_fields(
                self._app_name,
                self._model_name,
                self._pk,
                {self._field_name: {key: to_output}},
            )
        self._cache[key] = value

    def __len__(self) -> int:
        """Return the number of known fields."""
        if self._fields is not None:
            return len(self._fields)
        else:
            return len(self._cache)

    def __delitem__(self, name):
        """Delete the item from JSON field."""
        raise NotImplementedError("Delete is not implemented")

    def __iter__(self):
        """Return iterator."""
        return iter(self._cache)

    def __str__(self):
        """Return string representation."""
        return f"JSONDescriptor({self._model}, {self._field_name})"


class ModelMetaclass(type):
    """Construct fields for the Django class."""

    def __new__(mcs, name, bases, namespace, **kwargs):
        """Create new Model class."""
        model: Type[Model] = type.__new__(mcs, name, bases, namespace)
        fields: Dict[str, Field] = {}
        if name != "Model" and communicator is not None:
            app_name = model._app_name
            fields_details = communicator.get_model_fields_details(app_name, name)
            for field_name in fields_details:
                field_type, required, related_model_name = fields_details[field_name]
                if field_type in FIELDS_MAP:
                    kwargs = {"required": required}
                    if field_type == "ForeignKey":
                        id_field = IntegerField()
                        id_field.contribute_to_class(model, fields, f"{field_name}_id")
                        setattr(model, f"{field_name}_id", id_field)
                        kwargs["full_model_name"] = related_model_name
                    if field_type == "ManyToManyField":
                        kwargs["inner"] = ModelField(related_model_name)
                    field_class = FIELDS_MAP[field_type]
                    field = field_class(**kwargs)
                    field.contribute_to_class(model, fields, field_name)
                    setattr(model, field_name, field)

            model.fields = fields
        return model


class JSONModelEncoder(json.JSONEncoder):
    """Encode Model to JSON.

    When instance of Model class is serialized to JSON it should be serialized
    to its id property.
    """

    def default(self, o: Any):
        """Override default method."""
        if isinstance(o, Model):
            return o.id
        return json.JSONEncoder.default(self, o)


class Model(metaclass=ModelMetaclass):
    """Base django model."""

    _app_name = "App name"
    _model_name = "Model name"
    # A default list of extra fields that are returned on filter call.
    _filter_response_fields = []

    def __init__(self, pk: int):
        """Initialize."""
        self._pk = pk
        self._cache: Dict[str, Any] = {"id": pk}

    @property
    def full_model_name(self):
        """Return the full model name."""
        return f"{self._app_name}.{self._model_name}"

    @classmethod
    def filter(cls, **filters: Dict[str, Any]) -> List["Model"]:
        """Return a list of all objects that fit criteria."""
        # Make sure attributes have 'id' in the first place.
        attributes = filters.pop("__fields", None)
        attributes = attributes or cls._filter_response_fields
        attributes = ["id"] + [
            attribute for attribute in attributes if attribute != "id"
        ]

        objects = communicator.filter_objects(
            cls._app_name, cls._model_name, filters, attributes
        )
        models = []
        for entry in objects:
            model = cls(entry[0])
            for field_name, value in zip(attributes[1:], entry[1:]):
                field = model.fields[field_name]
                model._cache[field_name] = field.clean(value)
            models.append(model)
        return models

    @classmethod
    def exists(cls, **filters: Dict[str, Any]) -> List[int]:
        """Check if objects that fit criteria exists.

        If no such object exists empty list is returned.
        Else list of ids that fit the criteria is returned.
        """
        return [
            e[0]
            for e in communicator.filter_objects(
                cls._app_name, cls._model_name, filters, ["id"]
            )
        ]

    @classmethod
    def get(cls, **filters: Dict[str, Any]) -> "Model":
        """Get a single model based on filters.

        :raises RuntimeError: when different than one objects match the given
            criteria.
        """
        pks = communicator.filter_objects(
            cls._app_name, cls._model_name, filters, ["id"]
        )
        if len(pks) > 1:
            raise RuntimeError(
                f"Exactly one object should match the given criteria, received {len(pks)}."
            )
        elif len(pks) == 0:
            raise RuntimeError(
                "No objects match the given criteria or no permission to read object."
            )
        return cls(pks[0][0])

    @classmethod
    def create(cls, **object_data: Dict[str, Any]) -> int:
        """Create object with the given data.

        If creation was successfull return its id.
        """
        mappings = []
        for field_name in object_data:
            field = cls.fields.get(field_name)
            if field is not None and field.get_field_type() == "model":
                mappings.append(
                    (field_name, f"{field_name}_id", object_data[field_name].id)
                )
        for old_name, new_name, new_value in mappings:
            del object_data[old_name]
            object_data[new_name] = new_value
        communicator.encoder = JSONModelEncoder
        return cls(
            communicator.create_object(cls._app_name, cls._model_name, object_data)
        )

    def __str__(self):
        """Return a string representation."""
        return f"{self._model_name}(pk={self._pk})"

    def _set_field_data(self, field: Field, value: Any):
        """Set the value of the field."""
        self._cache[field.name] = value
        communicator.update_model_fields(
            self._app_name,
            self._model_name,
            self._pk,
            {field.name: field.to_output(value)},
        )

    def _get_field_data(self, field: Field) -> Any:
        """Get data for the given field.

        It is used by the fields (descriptors) to retrieve the data from the
        server. The cache is used if applicable.

        :raises RuntimeError: when no data is received.
        """
        if field.name not in self._cache:
            result = communicator.get_model_fields(
                self._app_name, self._model_name, self._pk, [field.name]
            )
            if len(result) > 1:
                result = [e[field.name] for e in result]
            elif len(result) == 1:
                result = result[field.name]
            else:
                raise RuntimeError(
                    (
                        f"No data received for property '{field.name}' of "
                        f"model {self._model_name} with id {self._pk}. Check "
                        "property name and permissions."
                    )
                )
            self._cache[field.name] = field.clean(result)
        return self._cache[field.name]

    @classmethod
    def __init_subclass__(cls: Type["Model"], **kwargs):
        """Register class."""
        super().__init_subclass__(**kwargs)
        RegisteredModels().register_model(cls)


class RegisteredModels:
    """Registered Python process models."""

    #  A single instance of this class.
    __instance = None
    _known_models: Dict[str, Type["Model"]] = dict()

    @classmethod
    def __new__(cls, *args, **kwargs):
        """Create or return singleton."""
        if RegisteredModels.__instance is None:
            RegisteredModels.__instance = super().__new__(cls)

        return RegisteredModels.__instance

    def get_model(self, full_model_name: str) -> Type["Model"]:
        """Get the registered model."""
        return self._known_models.get(full_model_name, Model)

    def register_model(self, model_class: Type["Model"]):
        """Add new model class.

        :raises AssertionError: when model is already registered.
        """
        full_model_name = f"{model_class._app_name}.{model_class._model_name}"
        assert (
            full_model_name not in self._known_models
        ), f"Model named {full_model_name} already registered."
        self._known_models[full_model_name] = model_class


class Process(Model):
    """Process model."""

    _app_name = "flow"
    _model_name = "Process"


class Entity(Model):
    """Entity model."""

    _app_name = "flow"
    _model_name = "Entity"


class Storage(Model):
    """Storage model."""

    _app_name = "flow"
    _model_name = "Storage"


class User(Model):
    """User model."""

    _app_name = "auth"
    _model_name = "User"


class Collection(Model):
    """Collection model."""

    _app_name = "flow"
    _model_name = "Collection"

    @property
    def relations(self):
        """Get relation."""
        if "__relations" not in self._cache:
            self._cache["__relations"] = communicator.get_relations(self.id)
        return self._cache["__relations"]


class Data(Model):
    """Data object."""

    _app_name = "flow"
    _model_name = "Data"

    @classmethod
    def from_slug(cls, slug: str) -> "Data":
        """Get Data object from slug."""
        return Data(communicator.get_data_by_slug(slug))

    @property
    def type(self) -> str:
        """Get the process type for this data object.

        DEPRECATED: compatibility with old DataDescriptor class.
        """
        return self.process.type

    @property
    def entity_name(self):
        """Entity name.

        Causes error when entity is not defined.

        DEPRECATED: compatibility with old DataDescriptor class.
        """
        return self.entity.name

    @property
    def relations(self):
        """Get relations for this data object in its collection."""
        relations = set()
        for relation in self.collection.relations:
            for partition in relation["partitions"]:
                if partition["entity_id"] == self.entity.id:
                    relations.add(RelationDescriptor.from_dict(relation))
        return list(relations)



# Generated by Django 2.2.6 on 2019-10-28 10:49

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0040_remove_entity_descriptor_completed"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="collection",
            options={
                "default_permissions": (),
                "get_latest_by": "version",
                "permissions": (
                    ("view_collection", "Can view collection"),
                    ("edit_collection", "Can edit collection"),
                    ("share_collection", "Can share collection"),
                    ("owner_collection", "Is owner of the collection"),
                ),
            },
        ),
        migrations.AlterModelOptions(
            name="data",
            options={
                "default_permissions": (),
                "get_latest_by": "version",
                "permissions": (
                    ("view_data", "Can view data"),
                    ("edit_data", "Can edit data"),
                    ("share_data", "Can share data"),
                    ("owner_data", "Is owner of the data"),
                ),
            },
        ),
        migrations.AlterModelOptions(
            name="entity",
            options={
                "default_permissions": (),
                "get_latest_by": "version",
                "permissions": (
                    ("view_entity", "Can view entity"),
                    ("edit_entity", "Can edit entity"),
                    ("share_entity", "Can share entity"),
                    ("owner_entity", "Is owner of the entity"),
                ),
            },
        ),
    ]



""".. Ignore pydocstyle D400.

====================
Kubernetes Connector
====================

"""
import hashlib
import json
import logging
import os
import re
import time
from contextlib import suppress
from pathlib import Path
from typing import Any, Dict, Iterable, Tuple

import kubernetes

from django.conf import settings
from django.db.models import Sum
from django.db.models.functions import Coalesce

from resolwe.flow.executors import constants
from resolwe.flow.executors.prepare import BaseFlowExecutorPreparer
from resolwe.flow.executors.protocol import ExecutorFiles
from resolwe.flow.models import Data, DataDependency, Process
from resolwe.storage import settings as storage_settings
from resolwe.storage.connectors import connectors
from resolwe.storage.connectors.baseconnector import BaseStorageConnector
from resolwe.utils import BraceMessage as __

from .base import BaseConnector

# TODO: is this really needed?
# Limits of containers' access to memory. We set the limit to ensure
# processes are stable and do not get killed by OOM signal.
KUBERNETES_MEMORY_HARD_LIMIT_BUFFER = 2000

# Timeout (in seconds) to wait for response from kubernetes API.
KUBERNETES_TIMEOUT = 30

logger = logging.getLogger(__name__)


class Connector(BaseConnector):
    """Kubernetes-based connector for job execution."""

    def __init__(self):
        """Initialize."""
        self._initialize_variables()

    def _initialize_variables(self):
        """Init variables.

        This has to be done for every run since settings values may be overriden for tests.
        """
        self.kubernetes_namespace = getattr(settings, "KUBERNETES_SETTINGS", {}).get(
            "namespace", "default"
        )
        self.runtime_dir = Path(
            storage_settings.FLOW_VOLUMES["runtime"]["config"]["path"]
        )

    def _get_upload_dir(self) -> str:
        """Get upload path.

        : returns: the path of the first mountable connector for storage
            'upload'.

        :raises RuntimeError: if no applicable connector is found.
        """
        for connector in connectors.for_storage("upload"):
            if connector.mountable:
                return f"/upload_{connector.name}"
        raise RuntimeError("No mountable upload connector is defined.")

    def _prepare_environment(
        self, data: Data, listener_connection: Tuple[str, str, str]
    ) -> list:
        """Prepare environmental variables."""
        host, port, protocol = listener_connection
        environment = {
            "LISTENER_IP": host,
            "LISTENER_PORT": port,
            "LISTENER_PROTOCOL": protocol,
            "DATA_ID": data.id,
            "FLOW_MANAGER_KEEP_DATA": getattr(
                settings, "FLOW_MANAGER_KEEP_DATA", False
            ),
            "RUNNING_IN_CONTAINER": 1,
            "RUNNING_IN_KUBERNETES": 1,
            "GENIALIS_UID": os.getuid(),
            "GENIALIS_GID": os.getgid(),
            "DESCRIPTOR_CHUNK_SIZE": 100,
            "MOUNTED_CONNECTORS": ",".join(
                connector.name
                for connector in connectors.values()
                if connector.mountable
            ),
        }
        with suppress(RuntimeError):
            environment["UPLOAD_DIR"] = self._get_upload_dir()

        return [
            {"name": name, "value": str(value)} for name, value in environment.items()
        ]

    def _get_upload_connector_name(self, data: Data) -> str:
        """Get the connector for data upload.

        Read the connector name from the StorageLocation class created by the
        dispatcher.
        """
        return data.location.default_storage_location.connector_name

    def _ebs_claim_name(self, base_name: str, data_id: int) -> str:
        """Get unique EBS claim name."""
        return f"{base_name}-{data_id}"

    def _get_files_configmap_name(self, location_subpath: Path, core_api: Any):
        """Get or create configmap for files.

        We have to map group, passwd and startup files inside containers.
        This is done using configmap which is mostly static. This command
        returns description for this configmap.
        """
        socket_utils_path: Path = (
            self.runtime_dir / "executors" / ExecutorFiles.SOCKET_UTILS
        )
        processing_startup_path: Path = (
            self.runtime_dir / "executors" / ExecutorFiles.STARTUP_PROCESSING_SCRIPT
        )
        constants_path: Path = self.runtime_dir / "executors" / ExecutorFiles.CONSTANTS

        startup_content = processing_startup_path.read_text()
        socket_utils_content = socket_utils_path.read_text()
        constants_content = constants_path.read_text()
        passwd_content = "root:x:0:0:root:/root:/bin/bash\n"
        passwd_content += f"user:x:{os.getuid()}:{os.getgid()}:user:{os.fspath(constants.PROCESSING_VOLUME)}:/bin/bash\n"
        group_content = "root:x:0:\n"
        group_content += f"user:x:{os.getgid()}:user\n"

        data = {
            "passwd": passwd_content,
            "group": group_content,
            "startup-script": startup_content,
            "socket-utils": socket_utils_content,
            "constants": constants_content,
        }

        data_md5 = hashlib.md5(json.dumps(data, sort_keys=True).encode()).hexdigest()
        configmap_name = f"configmap-files-{data_md5}"

        logger.debug(f"Files configmap: {configmap_name}")

        try:
            core_api.read_namespaced_config_map(
                name=configmap_name, namespace=self.kubernetes_namespace
            )
        except kubernetes.client.rest.ApiException:
            # The configmap is not found, create one.
            configmap_description = {
                "apiVersion": "v1",
                "kind": "ConfigMap",
                "metadata": {"name": configmap_name},
                "data": data,
            }
            # The configmap might have already been created in the meantime.
            with suppress(kubernetes.client.rest.ApiException):
                core_api.create_namespaced_config_map(
                    body=configmap_description,
                    namespace=self.kubernetes_namespace,
                    _request_timeout=KUBERNETES_TIMEOUT,
                )

        return configmap_name

    def _get_mountable_connectors(self) -> Iterable[Tuple[str, BaseStorageConnector]]:
        """Iterate through all the storages and find mountable connectors.

        :returns: list of tuples (storage_name, connector).
        """
        return (
            (storage_name, connector)
            for storage_name in storage_settings.FLOW_STORAGE
            for connector in connectors.for_storage(storage_name)
            if connector.mountable
        )

    def _volumes(self, location_subpath: Path, core_api: Any) -> list:
        """Prepare all volumes."""

        def volume_from_config(volume):
            """Get configuration for kubernetes for given volume."""
            name = volume["config"]["name"]
            if volume["type"] == "ebs":
                return {
                    "name": name,
                    "persistentVolumeClaim": {"claimName": name},
                }
            elif volume["type"] == "host_path":
                return {
                    "name": name,
                    "hostPath": {"path": os.fspath(volume["config"]["path"])},
                }
            else:
                raise RuntimeError(f"Unsupported volume configuration: {volume}.")

        files_configmap_name = self._get_files_configmap_name(
            location_subpath, core_api
        )

        volumes = [
            {
                "name": "files-volume",
                "configMap": {"name": files_configmap_name},
            },
            {"name": "sockets-volume", "emptyDir": {}},
            {"name": "secrets-volume", "emptyDir": {}},
        ]
        volumes += [
            volume_from_config(volume)
            for volume in storage_settings.FLOW_VOLUMES.values()
        ]
        volumes += [
            {
                "name": connector.name,
                "hostPath": {"path": os.fspath(connector.config["path"])},
            }
            for storage_name, connector in self._get_mountable_connectors()
        ]
        return volumes

    def _init_container_mountpoints(self):
        """Prepare mountpoints for init container.

        Processing and input volume (if defined) and all mountable connectors
        are mounted inside container.
        """
        mount_points = [
            {
                "name": storage_settings.FLOW_VOLUMES["processing"]["config"]["name"],
                "mountPath": os.fspath(constants.PROCESSING_VOLUME),
                "readOnly": False,
            },
            {
                "name": "secrets-volume",
                "mountPath": os.fspath(constants.SECRETS_VOLUME),
                "readOnly": False,
            },
        ]
        if "input" in storage_settings.FLOW_VOLUMES:
            mount_points.append(
                {
                    "name": storage_settings.FLOW_VOLUMES["input"]["config"]["name"],
                    "mountPath": os.fspath(constants.INPUTS_VOLUME),
                    "readOnly": False,
                }
            )
        mount_points += [
            {
                "name": connector.name,
                "mountPath": f"/{storage_name}_{connector.name}",
                "readOnly": False,
            }
            for storage_name, connector in self._get_mountable_connectors()
        ]
        return mount_points

    def _communicator_mountpoints(self, location_subpath: Path) -> list:
        """Mountpoints for communicator container.

        Socket directory and mountable connectors are mounted inside.
        """
        mount_points = [
            {
                "name": "sockets-volume",
                "mountPath": os.fspath(constants.SOCKETS_VOLUME),
                "readOnly": False,
            },
            {
                "name": "secrets-volume",
                "mountPath": os.fspath(constants.SECRETS_VOLUME),
                "readOnly": False,
            },
        ]
        mount_points += [
            {
                "name": connector.name,
                "mountPath": f"/{storage_name}_{connector.name}",
                "readOnly": False,
            }
            for storage_name, connector in self._get_mountable_connectors()
        ]
        return mount_points

    def _processing_mountpoints(
        self, location_subpath: Path, execution_engine_name: str
    ):
        """Mountpoints for processing container.

        Processing and input volume (if defined) and all mountable connectors
        are mounted inside container. All except processing volume are mounted
        read-only.
        """
        mount_points = [
            {
                "name": storage_settings.FLOW_VOLUMES["processing"]["config"]["name"],
                "mountPath": os.fspath(constants.PROCESSING_VOLUME),
                "subPath": os.fspath(location_subpath),
                "readOnly": False,
            },
        ]
        if "input" in storage_settings.FLOW_VOLUMES:
            mount_points.append(
                {
                    "name": storage_settings.FLOW_VOLUMES["input"]["config"]["name"],
                    "mountPath": os.fspath(constants.INPUTS_VOLUME),
                    "readOnly": False,
                }
            )
        mount_points += [
            {
                "name": connector.name,
                "mountPath": f"/{storage_name}_{connector.name}",
                "readOnly": storage_name != "upload",
            }
            for storage_name, connector in self._get_mountable_connectors()
        ]

        mount_points += [
            {
                "name": "files-volume",
                "mountPath": "/etc/passwd",
                "subPath": "passwd",
            },
            {
                "name": "files-volume",
                "mountPath": "/etc/group",
                "subPath": "group",
            },
            {
                "name": "files-volume",
                "mountPath": "/socket_utils.py",
                "subPath": "socket-utils",
            },
            {
                "name": "files-volume",
                "mountPath": "/processing.py",
                "subPath": "startup-script",
            },
            {
                "name": "files-volume",
                "mountPath": "/constants.py",
                "subPath": "constants",
            },
            {
                "name": "sockets-volume",
                "mountPath": os.fspath(constants.SOCKETS_VOLUME),
            },
            {
                "name": "secrets-volume",
                "mountPath": os.fspath(constants.SECRETS_VOLUME),
                "readOnly": True,
            },
        ]

        if "tools" in storage_settings.FLOW_VOLUMES:
            # Create volumes for tools. In kubernetes all tools must be inside
            # volume of type persistent_volume or host_path.
            volume = storage_settings.FLOW_VOLUMES["tools"]
            preparer = BaseFlowExecutorPreparer()
            tools_paths = preparer.get_tools_paths()
            base_tools_path = volume["config"]["path"]
            subpath = volume["config"].get("subpath", "")
            if subpath:
                tools_paths = [f"{subpath}/{path}" for path in tools_paths]
            mount_points += [
                {
                    "name": volume["config"]["name"],
                    "mountPath": os.fspath(Path("/usr/local/bin/resolwe") / str(index)),
                    "subPath": os.fspath(Path(tool).relative_to(base_tools_path)),
                }
                for index, tool in enumerate(tools_paths)
            ]

        from resolwe.flow.managers import manager

        execution_engine = manager.get_execution_engine(execution_engine_name)
        runtime_volumes = execution_engine.prepare_volumes()
        subpath = storage_settings.FLOW_VOLUMES["runtime"]["config"].get("subpath", "")
        if subpath:
            runtime_volumes = {
                f"{subpath}/{src}": dst for src, dst in runtime_volumes.items()
            }

        mount_points += [
            {
                "name": storage_settings.FLOW_VOLUMES["runtime"]["config"]["name"],
                "mountPath": dst,
                "subPath": src,
            }
            for src, dst in runtime_volumes.items()
        ]
        return mount_points

    def _persistent_ebs_claim(
        self, claim_name: str, volume_config: Dict, size: int
    ) -> Dict[str, Any]:
        # self, persistent_claim_name: str, volume_size_in_bytes: int
        """Prepare claim for EBS Amazon storage."""
        return {
            "apiVersion": "v1",
            "kind": "PersistentVolumeClaim",
            "metadata": {"name": claim_name},
            "spec": {
                "accessModes": ["ReadWriteOnce"],
                "storageClassName": volume_config.get("storageClassName", "gp3"),
                "resources": {
                    "requests": {
                        "storage": size,
                    }
                },
            },
        }

    def _data_inputs_size(self, data: Data, safety_buffer: int = 2 ** 30) -> int:
        """Get the size of data inputs.

        Also add 10% of the volume size + 2GB as safety buffer. When having a
        large filesystem with lots of files the filesystem overhead can be
        significant so a fixed safety buffer alone is not sufficient.

        :returns: the size of the input data in bytes.
        """
        inputs_size = (
            Data.objects.filter(
                children_dependency__child=data,
                children_dependency__kind=DataDependency.KIND_IO,
            )
            .aggregate(total_size=Coalesce(Sum("size"), 0))
            .get("total_size")
        )
        assert isinstance(inputs_size, int)
        return int(1.1 * (inputs_size + safety_buffer))

    def _sanitize_kubernetes_label(self, label: str, trim_end: bool = True) -> str:
        """Make sure kubernetes label complies with the rules.

        See the URL bellow for details.

        https://kubernetes.io/docs/concepts/overview/working-with-objects/labels/
        """
        max_length = 63
        sanitized_label = re.sub("[^0-9a-zA-Z\-]+", "-", label).strip("-_.")
        if len(sanitized_label) > max_length:
            logger.warning(__("Label '%s' is too long and was truncated.", label))
            if trim_end:
                sanitized_label = sanitized_label[-max_length:].strip("-_.")
            else:
                sanitized_label = sanitized_label[:max_length].strip("-_.")
        return sanitized_label

    def start(self, data: Data, listener_connection: Tuple[str, str, str]):
        """Start process execution.

        Construct kubernetes job description and pass it to the kubernetes.
        """
        container_environment = self._prepare_environment(data, listener_connection)

        location_subpath = Path(data.location.subpath)

        # Create kubernetes API every time otherwise it will time out
        # eventually and raise API exception.
        kubernetes.config.load_kube_config()
        batch_api = kubernetes.client.BatchV1Api()
        core_api = kubernetes.client.CoreV1Api()

        container_name_prefix = (
            getattr(settings, "FLOW_EXECUTOR", {})
            .get("CONTAINER_NAME_PREFIX", "resolwe")
            .replace("_", "-")
            .lower()
        )
        container_name = self._generate_container_name(container_name_prefix, data.pk)

        # Set resource limits.
        requests = dict()
        limits = data.process.get_resource_limits()

        requests["cpu"] = limits.pop("cores")
        limits["cpu"] = requests["cpu"] + 1
        # Overcommit CPU by 20%.
        requests["cpu"] *= 0.8

        # The memory in the database is stored in megabytes but the kubertenes
        # requires memory in bytes.
        # We request 10% less memory than stored in the database and set limit
        # at 10% more plus KUBERNETES_MEMORY_HARD_LIMIT_BUFFER. The processes
        # usually require 16GB, 32GB... and since the node usualy has 64GB of
        # memory and some of it is consumed by the system processes only one
        # process process that requires 32GB can run on a node instead of 2.

        requests["memory"] = 0.9 * limits["memory"]
        limits["memory"] = 1.1 * limits["memory"] + KUBERNETES_MEMORY_HARD_LIMIT_BUFFER
        limits["memory"] *= 2 ** 20  # 2 ** 20 = mebibyte
        requests["memory"] *= 2 ** 20

        resources = data.process.requirements.get("resources", {})
        network = "bridge"
        use_host_network = False
        if "network" in resources:
            # Configure Docker network mode for the container (if specified).
            # By default, current Docker versions use the 'bridge' mode which
            # creates a network stack on the default Docker bridge.
            network = getattr(settings, "FLOW_EXECUTOR", {}).get("NETWORK", "")
            use_host_network = network == "host"

        # Generate and set seccomp policy to limit syscalls.
        security_context = {
            "runAsUser": os.getuid(),
            "runAsGroup": os.getgid(),
            "allowPrivilegeEscalation": False,
            "privileged": False,
            "capabilities": {"drop": ["ALL"]},
        }

        annotations = dict()

        # Do not evict job from node.
        annotations["cluster-autoscaler.kubernetes.io/safe-to-evict"] = "false"

        if not getattr(settings, "FLOW_DOCKER_DISABLE_SECCOMP", False):
            # The path is a relative path in the kubelet root
            # directory:
            # <seccomp_root>/<path>, where <seccomp_root> is defined via the
            # --seccomp-profile-root flag on the Kubelet. If the
            # --seccomp-profile-root flag is not defined, the default path will
            # be used, which is <root-dir>/seccomp where <root-dir> is
            # specified by the --root-dir flag.
            # https://kubernetes.io/docs/concepts/policy/pod-security-policy/
            #
            # The file is transfered to kubelets with daemonset ? Currently I
            # mount my /tmp directory to the /seccomp directory in minikube.
            annotations["seccomp.security.alpha.kubernetes.io/pod"] = "runtime/default"

        communicator_image = getattr(
            settings,
            "FLOW_DOCKER_COMMUNICATOR_IMAGE",
            "public.ecr.aws/s4q6j6e8/resolwe/com:latest",
        )

        requirements = data.process.requirements.get("executor", {}).get("docker", {})
        processing_container_image = str(
            requirements.get(
                "image",
                getattr(
                    settings,
                    "FLOW_DOCKER_DEFAULT_PROCESSING_CONTAINER_IMAGE",
                    "public.ecr.aws/s4q6j6e8/resolwe/base:ubuntu-20.04",
                ),
            ),
        )

        job_type = dict(Process.SCHEDULING_CLASS_CHOICES)[data.process.scheduling_class]
        job_description = {
            "apiVersion": "batch/v1",
            "kind": "Job",
            "metadata": {"name": self._sanitize_kubernetes_label(container_name)},
            "spec": {
                # Keep finished pods around for ten seconds. If job is not
                # deleted its PVC claim persists and it causes PV to stay0
                # around.
                # This can be changed by running a cron job that periodically
                # checks for PVC that can be deleted.
                "ttlSecondsAfterFinished": 10,
                "template": {
                    "metadata": {
                        "name": self._sanitize_kubernetes_label(container_name),
                        "labels": {
                            "app": "resolwe",
                            "data_id": str(data.pk),
                            "process": self._sanitize_kubernetes_label(
                                data.process.slug
                            ),
                            "job_type": self._sanitize_kubernetes_label(job_type),
                        },
                        "annotations": annotations,
                    },
                    "spec": {
                        "hostNetwork": use_host_network,
                        "volumes": self._volumes(location_subpath, core_api),
                        "initContainers": [
                            {
                                "name": self._sanitize_kubernetes_label(
                                    f"{container_name}-init"
                                ),
                                "image": communicator_image,
                                "imagePullPolicy": "Always",
                                "workingDir": "/",
                                "command": ["/usr/local/bin/python3"],
                                "args": ["-m", "executors.init_container"],
                                "securityContext": {"privileged": True},
                                "volumeMounts": self._init_container_mountpoints(),
                                "env": container_environment,
                            },
                        ],
                        "containers": [
                            {
                                "name": self._sanitize_kubernetes_label(container_name),
                                "image": processing_container_image,
                                "resources": {"limits": limits, "requests": requests},
                                "securityContext": security_context,
                                "env": container_environment,
                                "workingDir": os.fspath(constants.PROCESSING_VOLUME),
                                "imagePullPolicy": "Always",
                                "command": ["/usr/bin/python3"],
                                "args": ["/processing.py"],
                                "volumeMounts": self._processing_mountpoints(
                                    location_subpath,
                                    data.process.run.get("language", None),
                                ),
                            },
                            {
                                "name": self._sanitize_kubernetes_label(
                                    f"{container_name}-communicator"
                                ),
                                "image": communicator_image,
                                "imagePullPolicy": "Always",
                                "resources": {
                                    "limits": {"cpu": 1, "memory": "1024M"},
                                    "requests": {"memory": "256M", "cpu": 0.1},
                                },
                                "securityContext": security_context,
                                "env": container_environment,
                                "command": ["/usr/local/bin/python3"],
                                "args": ["/startup.py"],
                                "volumeMounts": self._communicator_mountpoints(
                                    location_subpath
                                ),
                            },
                        ],
                        "restartPolicy": "Never",
                    },
                },
                "backoffLimit": 0,
            },
        }
        start_time = time.time()
        limits.pop("storage", 200)

        if storage_settings.FLOW_VOLUMES["processing"]["type"] == "ebs":
            claim_name = self._ebs_claim_name(
                storage_settings.FLOW_VOLUMES["processing"]["config"]["name"], data.id
            )
            claim_size = limits.pop("storage", 200) * (2 ** 30)  # Default 200 gibibytes
            core_api.create_namespaced_persistent_volume_claim(
                body=self._persistent_ebs_claim(
                    claim_name,
                    claim_size,
                    storage_settings.FLOW_VOLUMES["processing"]["config"],
                ),
                namespace=self.kubernetes_namespace,
                _request_timeout=KUBERNETES_TIMEOUT,
            )
        if hasattr(settings, "FLOW_INPUTS_VOLUME"):
            if settings.FLOW_INPUTS_VOLUME["type"] == "ebs":
                claim_size = self._data_inputs_size(data)
                claim_name = self._ebs_claim_name(
                    settings.FLOW_INPUTS_VOLUME["config"]["name"], data.id
                )
                core_api.create_namespaced_persistent_volume_claim(
                    body=self._persistent_ebs_claim(
                        claim_name, claim_size, settings.FLOW_INPUTS_VOLUME["config"]
                    ),
                    namespace=self.kubernetes_namespace,
                    _request_timeout=KUBERNETES_TIMEOUT,
                )

        batch_api.create_namespaced_job(
            body=job_description,
            namespace=self.kubernetes_namespace,
            _request_timeout=KUBERNETES_TIMEOUT,
        )
        end_time = time.time()
        logger.info(
            "It took {:.2f}s to send config to kubernetes".format(end_time - start_time)
        )

    def _generate_container_name(self, prefix: str, data_id: int):
        """Generate unique container name.

        Name of the kubernetes container should contain only lower case
        alpfanumeric characters and dashes. Underscores are not allowed.
        """
        return "{}-{}".format(prefix, data_id)

    def submit(self, data: Data, runtime_dir: str, argv: Any):
        """Run process.

        For details, see
        :meth:`~resolwe.flow.managers.workload_connectors.base.BaseConnector.submit`.
        """
        (host, port, protocol) = argv[-1].rsplit(" ", maxsplit=3)[-3:]
        self._initialize_variables()
        self.start(data, (host, port, protocol))

        logger.debug(
            __(
                "Connector '{}' running for Data with id {} ({}).",
                self.__class__.__module__,
                data.id,
                repr(argv),
            )
        )

    def cleanup(self, data_id: int):
        """Remove the EBS storage used by the executor."""
        kubernetes.config.load_kube_config()
        core_api = kubernetes.client.CoreV1Api()
        claim_names = [
            self._ebs_claim_name(type_, data_id) for type_ in ["local", "inputs"]
        ]
        for ebs_claim_name in claim_names:
            logger.debug("Kubernetes: removing claim %s.", ebs_claim_name)
            with suppress(kubernetes.client.rest.ApiException):
                core_api.delete_namespaced_persistent_volume_claim(
                    name=ebs_claim_name,
                    namespace=self.kubernetes_namespace,
                    _request_timeout=KUBERNETES_TIMEOUT,
                )



"""Utility methods for receiving/sending data over socket."""
import asyncio
import functools
import json
import logging
import random
import socket
import time
import uuid
from collections import deque
from contextlib import suppress
from enum import Enum, unique
from time import time as now
from typing import (
    Any,
    Callable,
    Coroutine,
    Deque,
    Dict,
    Generic,
    MutableMapping,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
)

logger = logging.getLogger(__name__)


@unique
class MessageType(Enum):
    """Type of the message."""

    COMMAND = "COMMAND"
    RESPONSE = "RESPONSE"
    HEARTBEAT = "HBT"
    HEARTBEAT_RESPONSE = "HBTR"
    ENQUIRE = "EQR"


class ResponseStatus(Enum):
    """Response status."""

    OK = "OK"
    ERROR = "ERR"


class PeerStatus(Enum):
    """Peer status."""

    RESPONSIVE = "responsive"
    DEGRADED = "degraded"
    UNRESPONSIVE = "unresponsive"


PeerIdentity = bytes
MessageDataType = TypeVar("MessageDataType")
ResponseDataType = TypeVar("ResponseDataType")


def retry(
    max_retries: int = 3,
    retry_exceptions: Tuple[Type[Exception]] = (ConnectionError,),
    min_sleep: int = 1,
    max_sleep: int = 10,
):
    """Try to call decorated method max_retries times before giving up.

    The calls are retried when function raises exception in retry_exceptions.

    :param max_retries: maximal number of calls before giving up.
    :param retry_exceptions: retry call if one of these exceptions is raised.
    :param min_sleep: minimal sleep between calls (in seconds).
    :param max_sleep: maximal sleep between calls (in seconds).
    :returns: return value of the called method.
    :raises: the last exceptions raised by the method call if none of the
      retries were successfull.
    """

    def decorator_retry(func):
        @functools.wraps(func)
        def wrapper_retry(*args, **kwargs):
            last_error: Exception = Exception("Retry failed")
            sleep: int = 0
            for retry in range(max_retries):
                try:
                    time.sleep(sleep)
                    return func(*args, **kwargs)
                except retry_exceptions as err:
                    sleep = min(max_sleep, min_sleep * (2 ** retry))
                    last_error = err
            raise last_error

        return wrapper_retry

    return decorator_retry


def async_retry(
    max_retries: int = 3,
    retry_exceptions: Tuple[Type[Exception]] = (ConnectionError,),
    min_sleep: int = 1,
    max_sleep: int = 10,
):
    """Try to call decorated method max_retries times before giving up.

    The calls are retried when function raises exception in retry_exceptions.

    :param max_retries: maximal number of calls before giving up.
    :param retry_exceptions: retry call if one of these exceptions is raised.
    :param min_sleep: minimal sleep between calls (in seconds).
    :param max_sleep: maximal sleep between calls (in seconds).
    :returns: return value of the called method.
    :raises: the last exceptions raised by the method call if none of the
      retries were successfull.
    """

    def decorator_retry(func):
        @functools.wraps(func)
        async def wrapper_retry(*args, **kwargs):
            last_error: Exception = Exception("Retry failed")
            sleep: int = 0
            for retry in range(max_retries):
                try:
                    await asyncio.sleep(sleep)
                    return await func(*args, **kwargs)
                except retry_exceptions as err:
                    sleep = min(max_sleep, min_sleep * (2 ** retry))
                    last_error = err
            raise last_error

        return wrapper_retry

    return decorator_retry


def send_data(
    s: socket.SocketType,
    data: dict,
    size_bytes: int = 8,
    encoder: Optional[Type[json.JSONEncoder]] = None,
):
    """Send data over socket.

    :param s: socket to send over.
    :param data: dict that must be serialzable to JSON.
    :param size_bytes: how first many bytes in message are dedicated to the
        message size (pre-padded with zeros).
    :raises: exception on failure.
    """
    message = json.dumps(data, cls=encoder).encode()
    message_length = len(message).to_bytes(size_bytes, byteorder="big")
    s.sendall(message_length)
    s.sendall(message)


def read_bytes(s: socket.SocketType, message_size: int) -> bytes:
    """Read message_size bytes from the given socket.

    The method will block until enough bytes are available.

    :param message_size: size (in bytes) of the message to read.

    :returns: received message.
    """
    message = b""
    while len(message) < message_size:
        received = s.recv(message_size - len(message))
        message += received
        if not received:
            return message
    return message


def receive_data(s: socket.SocketType, size_bytes: int = 8) -> Optional[Dict]:
    """Recieve data over the given socket.

    :param size_bytes: how first many bytes in message are dedicated to the
        message size (pre-padded with zeros).
    """
    message = read_bytes(s, size_bytes)
    if not message:
        return None
    message_size = int.from_bytes(message, byteorder="big")

    message = read_bytes(s, message_size)
    assert len(message) == message_size
    data = json.loads(message.decode("utf-8"))
    return data


async def async_send_data(
    writer: asyncio.StreamWriter,
    data: dict,
    identity: Optional[bytes] = None,
    size_bytes: int = 8,
):
    """Send data over socket.

    :param writer: object to write into.
    :param data: JSON serializable object.
    :param identity: ignored by socker writer.
    :param size_bytes: how first many bytes in message are dedicated to the
        message size (pre-padded with zeros).
    :raises: exception on failure.
    """
    message = json.dumps(data).encode("utf-8")
    message_size = len(message).to_bytes(size_bytes, byteorder="big")
    writer.write(message_size)
    writer.write(message)
    await writer.drain()


async def async_receive_data(
    reader: asyncio.StreamReader, size_bytes: int = 8
) -> Optional[Tuple[PeerIdentity, Any]]:
    """Receive data from the reader.

    The data is expected to be bytes-encoded JSON representation of a Python
    object. Received data is deserialized to a Python object.

    :raises asyncio.IncompleteReadError: when data could not be read from
        the socket.

    :returns: optional tuple where first element is identity of the messenger
        (None with socket) and the second tuple is the received message.
    """
    received = await reader.readexactly(size_bytes)
    message_size = int.from_bytes(received, byteorder="big")
    received = await reader.readexactly(message_size)
    assert len(received) == message_size
    return (b"", json.loads(received.decode("utf-8")))


class Message(Generic[MessageDataType]):
    """Base class for message."""

    def __init__(
        self,
        message_type: MessageType,
        type_data: str,
        message_data: MessageDataType,
        message_uuid: Optional[str] = None,
        sequence_number: Optional[int] = None,
        sent_timestamp: Optional[float] = None,
    ):
        """Initialize.

        The sent_timestamp is auto-set on sending message to timestamp.
        """
        self.message_type = message_type
        self.message_data = message_data
        self.type_data = type_data
        self.uuid = message_uuid or self._get_random_message_identifier()
        self.sequence_number = sequence_number or 1
        self.sent_timestamp = sent_timestamp

    def _get_random_message_identifier(self) -> str:
        """Get a random message identifier.

        This identifier is uses to match a response with the request.
        """
        return uuid.uuid4().hex

    @property
    def response_status(self) -> ResponseStatus:
        """Get response status if message type is ``MessageType.RESPONSE``.

        :raises ValueError: when type_data is not one of value in
            ResponseStatus enum.

        :raises AssertionError: if type is not ``MessageType.RESPONSE``.
        """
        assert self.message_type is MessageType.RESPONSE
        return ResponseStatus(self.type_data)

    @property
    def command_name(self) -> str:
        """Get response status if message type is ``MessageType.RESPONSE``.

        :raises AssertionError: if type is not ``MessageType.RESPONSE``.
        """
        assert self.message_type is MessageType.COMMAND
        return self.type_data

    def respond(
        self,
        response_data: ResponseDataType,
        response_status: ResponseStatus = ResponseStatus.OK,
    ) -> "Response[ResponseDataType]":
        """Return a response with the given data and status."""
        return Response(
            response_status.value,
            response_data,
            self.uuid,
            self.sequence_number,
        )

    def respond_heartbeat(self) -> "Response[str]":
        """Respond heartbeat messages."""
        response = self.respond_ok("")
        response.message_type = MessageType.HEARTBEAT_RESPONSE
        return response

    def respond_ok(
        self, response_data: ResponseDataType
    ) -> "Response[ResponseDataType]":
        """Respond OK with the given data."""
        return self.respond(response_data, ResponseStatus.OK)

    def respond_error(
        self, response_data: ResponseDataType
    ) -> "Response[ResponseDataType]":
        """Respond ERROR with the given data."""
        return self.respond(response_data, ResponseStatus.ERROR)

    @staticmethod
    def command(
        command_name: str,
        message_data: MessageDataType,
        message_uuid: Optional[str] = None,
        sequence_number: Optional[int] = None,
    ) -> "Message[MessageDataType]":
        """Construct and return a command."""
        return Message(
            MessageType.COMMAND,
            command_name,
            message_data,
            message_uuid,
            sequence_number,
        )

    @staticmethod
    def heartbeat() -> "Message[str]":
        """Construct the heartbeat message."""
        return Message(MessageType.HEARTBEAT, "", "")

    @staticmethod
    def is_valid(message_dict: Dict) -> bool:
        """Validate the dictionary representing the message."""
        try:
            required_keys = ("type", "type_data", "data", "uuid", "sequence_number")
            value_types: Tuple = (str, str, object, str, int)
            assert set(required_keys).issubset(set(message_dict.keys()))
            assert all(
                isinstance(message_dict[key], value_type)
                for key, value_type in zip(required_keys, value_types)
            )
            message_type = MessageType(message_dict["type"])
            # If message is response chech that type_data is of the correct type.
            if message_type is MessageType.RESPONSE:
                ResponseStatus(message_dict["type_data"])
        except (AssertionError, ValueError):
            return False
        else:
            return True

    @staticmethod
    def from_dict(message_dict: Dict) -> "Message":
        """Construct a Message from dictionary.

        :raises KeyError, ValueError: if dictionary does not represent the
            valid message.
        """
        if message_dict["type"] == MessageType.RESPONSE.value:
            return Response.from_dict(message_dict)
        return Message(
            MessageType(message_dict["type"]),
            message_dict["type_data"],
            message_dict["data"],
            message_dict.get("uuid"),
            message_dict.get("sequence_number"),
            message_dict.get("timestamp"),
        )

    def to_dict(self) -> dict:
        """Get representation of the message as a dictionary.

        It is suitable for serialization into JSON.
        """
        return {
            "type": self.message_type.value,
            "type_data": self.type_data,
            "data": self.message_data,
            "uuid": self.uuid,
            "sequence_number": self.sequence_number,
            "timestamp": self.sent_timestamp,
        }

    def __repr__(self) -> str:
        """Return the string representation."""
        return str(self.to_dict())

    def __eq__(self, other):
        """Equality check."""
        if isinstance(other, Message):
            return (
                self.type_data == other.type_data
                and self.message_data == other.message_data
                and self.message_type == other.message_type
            )
        return False


class Response(Message[MessageDataType]):
    """Response message."""

    def __init__(
        self,
        type_data: str,
        message_data: MessageDataType,
        message_uuid: Optional[str] = None,
        sequence_number: Optional[int] = None,
        sent_timestamp: Optional[float] = None,
    ):
        """Initialize."""
        super().__init__(
            MessageType.RESPONSE,
            type_data,
            message_data,
            message_uuid,
            sequence_number,
            sent_timestamp,
        )
        self.status: ResponseStatus = ResponseStatus(type_data)

    @staticmethod
    def from_dict(message_dict: Dict) -> "Response":
        """Construct a Response from dictionary.

        :raises KeyError, ValueError: if dictionary does not represent the
            valid message.
        """
        assert message_dict["type"] == MessageType.RESPONSE.value
        return Response(
            message_dict["type_data"],
            message_dict["data"],
            message_dict.get("uuid"),
            message_dict.get("sequence_number"),
            message_dict.get("timestamp"),
        )


class MessageStatus(Enum):
    """Status of the message."""

    RECEIVED = "received"
    UNKNOWN = "unknown"


class EventWithResponse(asyncio.Event):
    """Event class with response property."""

    def __init__(self):
        """Initialize."""
        super().__init__()
        self._response: Optional[Any] = None

    @property
    def response(self) -> Any:
        """Return the response.

        :raises RuntimeError: when response is not set.
        """
        if self._response is None:
            raise RuntimeError("Response is not set")

        return self._response

    @response.setter
    def response(self, message_data: Any):
        self._response = message_data


class BaseCommunicator:
    """Handles socket communication.

    The communication is always iniciated by the command which must be followed
    by the response. The command is the JSON representation (encoded to bytes)
    of the dictionary of the following format::

        {
            "command": "command_name",
            "data": additional_command_data,
            "uuid": unique_command_id
        }

    The response to the command is also JSON representation (encoded to bytes)
    of the dictionary of the following format::

        {
            "response": "OK/ERROR",
            "data": additional_response_data,
            "uuid": unique_command_id
        }

    , where ``uuid`` in the response must match the one in the command.

    This class has two basic methods: one for sending commands and one for
    reponding to them. The method that receives commands must be run as a
    separate task.

    To cleanly shutdown the receiving task one must call a
    ``terminate`` method and wait for the receiving task to finish.
    """

    FINAL_COMMAND_NAME = "finish"

    def __init__(
        self,
        name: str,
        logger: logging.Logger,
        reader: Any,
        writer: Any,
        send_method: Callable,
        receive_method: Callable,
        peer_status_changed: Optional[
            Callable[[PeerIdentity, PeerStatus], Coroutine]
        ] = None,
    ):
        """Initialize.

        :param name: used in logging to identify the communicator.
        :param logger: logger to use for logging.
        :param reader: object to use when receiving messages.
        :param writer: object to use when sending messages.
        :param send_method: method to use when sending messages.
        :param receive_method: method to use when receiving messages.

        """
        self.name = name
        self.logger = logger
        self._terminating = asyncio.Event()
        self.has_message = asyncio.Event()
        self.send_method = send_method
        self.receive_method = receive_method
        self.reader = reader
        self.writer = writer

        self._listening_future: Optional[asyncio.Future] = None
        # Wait for x seconds for listening future to complete when _terminating
        # flag is set.
        self._listening_future_wait_timeout = 30

        self.sequence_numbers: MutableMapping[PeerIdentity, int] = dict()
        self._command_queue: Deque[Tuple[PeerIdentity, Message]] = deque()

        self._uuid_to_event: Dict[str, EventWithResponse] = dict()

        # Keep two lists of received messages. After 10 minutes move current
        # messages to older list and discard older messages. So the list
        # of received messages will not become too large.
        self._uuids_received: Set[str] = set()
        self._uuids_received_old: Set[str] = set()
        # When no message has been exchanged with the peer for
        # _heartbeat_interval seconds the heartbeat message is sent.
        self._heartbeat_interval = 60

        # After that many unanswered heartbeats the peer is removed from the
        # _known_peers list and heartbeats are no longer sent to it.
        self._max_heartbeats_skipped = 2
        # Mapping from peer identity to the timestamp of the last message
        # received by the peer.
        self._known_peers: Dict[PeerIdentity, float] = dict()
        self._degraded_peers: Set[PeerIdentity] = set()
        self._last_heartbeat: Dict[PeerIdentity, float] = dict()
        self.peer_status_changed = peer_status_changed
        self._heartbeat_messages_future: Optional[asyncio.Future] = None
        self._watchdog_future: Optional[asyncio.Future] = None
        self._recycle_uuids_future: Optional[asyncio.Future] = None

    def __getattr__(self, name: str):
        """Call arbitrary 'command' with 'communicator.command(args)' syntax."""

        def call_command(*args):
            if len(args) == 1:
                args = args[0]
            return self.send_command(Message.command(name, args))

        if name.startswith("_"):
            return None
        else:
            return call_command

    async def _recycle_received_uuids(self, timeout: int = 600):
        """Recycle the list of received messages uuids.

        Move received messages to the old list and discard old ones.
        """
        while True:
            self._uuids_received_old = self._uuids_received
            self._uuids_received = set()
            await asyncio.sleep(timeout)

    async def _receive_message(self) -> Optional[Tuple[PeerIdentity, Message]]:
        """Receive a single message.

        This method is blocking: it waits for the message to arrive. The
        method return only when one of the following happens:

        1. The message is received. In this case the message is returned.

        2. The terminating flag is set. In this case None is returned.

        3. Unexpected error occurs when receiving message. In this case None
           is returned.

        :returns: received message or None if message could not be received.
            The returned message is guaranteed to be dictionary and contain
            the key "type".
        """
        result = None
        try:
            self.logger.debug("Communicator %s waiting for message.", self.name)
            receive_task = asyncio.ensure_future(self.receive_method(self.reader))
            terminating_task = asyncio.ensure_future(self._terminating.wait())
            done, pending = await asyncio.wait(
                (receive_task, terminating_task),
                return_when=asyncio.FIRST_COMPLETED,
            )
            if receive_task in done:
                try:
                    received = receive_task.result()
                except asyncio.exceptions.IncompleteReadError:
                    self.logger.info("Socket closed by peer, stopping communication.")
                    received = None
                if received is not None:
                    assert isinstance(received, tuple)
                    assert len(received) == 2
                    assert isinstance(received[0], bytes)
                    assert Message.is_valid(received[1])
                    result = received[0], Message.from_dict(received[1])
            else:
                self.logger.debug(
                    "Communicator %s _receive_message: terminating flag is set, returning None",
                    self.name,
                )
        except:
            self.logger.exception(
                "Communicator %s: exception in _receive_message.", self.name
            )
        finally:
            # Always stop both tasks.
            receive_task.cancel()
            terminating_task.cancel()
        return result

    async def _send_message(
        self,
        message: Message,
        identity: PeerIdentity = b"",
        send_timeout: Optional[float] = 6,
        send_retries: int = 10,
    ) -> Any:
        """Send message using the supplied method.

        If message does not contains key "uuid" one is chosen at random.

        :param message: the message to send.

        :param identity: the identity of the peer to send message to.

        :param send_timeout: timeout (in seconds) to wait for the reply. If
            its value is None the timeout is infinite. This timeout will
            probably never occur since 0mq has internal buffer for sending
            messages.

        :param send_retries: try to resend command this many times if error
            occurs when sending.

        :returns: the result of the sending method call (usually None).

        :raises RuntimeError: when user supplied method raises exception when
            sending message or timeout occured when sending message.
        """
        retries = 0
        while retries < send_retries:
            try:
                message.sent_timestamp = now()
                data_to_send = message.to_dict()
                self.logger.debug(
                    "Communicator %s sending data: %s.", self.name, data_to_send
                )

                return await asyncio.wait_for(
                    self.send_method(self.writer, message.to_dict(), identity),
                    timeout=send_timeout,
                )
            except asyncio.TimeoutError:
                self.logger.error(
                    f"Communicator {self.name}: sending message timeout (retry {retries+1}/{send_retries})."
                )
            except asyncio.CancelledError:
                self.logger.info(
                    f"Communicator {self.name}: sending message was canceled."
                )
                return
            except Exception:
                self.logger.exception(
                    f"Communicator {self.name}: exception while sending message (retry {retries+1}/{send_retries})."
                )
            finally:
                retries += 1
        raise RuntimeError(f"Sending message ({message}, {str(identity)}) timeout.")

    async def __aenter__(self):
        """Start listening for messages on entering context."""
        assert self._listening_future is None
        assert self._heartbeat_messages_future is None
        assert self._watchdog_future is None
        assert self._recycle_uuids_future is None

        self._terminating.clear()

        self._listening_future = asyncio.ensure_future(self.start_listening())
        self._heartbeat_messages_future = asyncio.ensure_future(
            self._send_heartbeat_messages()
        )
        self._watchdog_future = asyncio.ensure_future(self._watchdog())
        self._recycle_uuids_future = asyncio.ensure_future(
            self._recycle_received_uuids()
        )

        self.logger.debug("Communicator %s: entering context.", self.name)
        return self

    async def __aexit__(self, typ, value, trace):
        """On exiting a context, stop listening for messages."""
        self.logger.debug(f"Communicator {self.name}: leaving context.")
        assert self._heartbeat_messages_future is not None
        assert self._watchdog_future is not None
        assert self._listening_future is not None
        assert self._recycle_uuids_future is not None

        self._terminating.set()
        with suppress(asyncio.TimeoutError):
            await asyncio.wait_for(
                self._listening_future, timeout=self._listening_future_wait_timeout
            )
        self._watchdog_future.cancel()
        self._heartbeat_messages_future.cancel()
        self._recycle_uuids_future.cancel()
        self._watchdog_future = None
        self._listening_future = None
        self._listening_future = None
        self.logger.debug("Communicator %s: leaving context.", self.name)

    def _message_status(self, message_uuid: str) -> MessageStatus:
        """Get status of message with the given UUID."""
        self.logger.debug("Checking for message status: %s.", message_uuid)
        if (
            message_uuid in self._uuids_received
            or message_uuid in self._uuids_received_old
        ):
            return MessageStatus.RECEIVED
        else:
            return MessageStatus.UNKNOWN

    async def _status_changed(self, peer: PeerIdentity, status: PeerStatus):
        """Notify (if possible) of the peer status change."""
        if self.peer_status_changed is not None:
            try:
                await self.peer_status_changed(peer, status)
            except:
                self.logger.exception("Exception in peer_status_changed.")

    async def _get_peer_message_status(
        self, message_uuid: str, identity: bytes = b""
    ) -> str:
        """Enquire peer for the message status.

        :raises RuntimeError: on failure.
        """
        result = await self.send_command(
            Message(MessageType.ENQUIRE, "", message_uuid),
            peer_identity=identity,
            response_timeout=60,
        )
        return result.message_data

    async def _send_heartbeat_messages(self):
        """Send heartbeat messages when necessary.

        This coroutine never stops and must be cancelled by the parent.
        """
        while True:
            try:
                self.logger.debug(
                    "Heartbeat checking known peers %s.", list(self._known_peers.keys())
                )
                # Wait a little longer than heartbeat interval if there is nothing to do.
                min_sleep_interval = self._heartbeat_interval + 5
                for identity, last_received in self._known_peers.items():
                    # Count time from last received message or last sent heartbeat.
                    last_event = max(
                        last_received, self._last_heartbeat.get(identity, last_received)
                    )
                    logger.debug(
                        "Last seen %s: %d seconds ago",
                        identity,
                        now() - last_received,
                    )
                    # Add a little random noise to spread the heartbeats around.
                    sleep_interval = int(
                        last_event
                        + self._heartbeat_interval
                        + random.uniform(0, 5 + 1)
                        - now()
                    )
                    logger.debug("Heartbeat in: %d seconds.", sleep_interval)

                    if sleep_interval <= 0:
                        asyncio.ensure_future(
                            self._send_message(Message.heartbeat(), identity=identity)
                        )
                        self._last_heartbeat[identity] = now()
                    else:
                        min_sleep_interval = min(min_sleep_interval, sleep_interval)
                self.logger.debug(
                    "Next heartbeat check: in %d seconds.", min_sleep_interval
                )
                await asyncio.sleep(min_sleep_interval)
            except Exception:
                self.logger.exception("Unexpected exception while sending heartbeats.")

    def suspend_heartbeat(self, peer_identity: PeerIdentity):
        """Temporary suspend heartbeat from the peer.

        The heartbeats are resumed on the first message received from the peer.
        """
        self._last_heartbeat.pop(peer_identity, None)
        self._known_peers.pop(peer_identity, None)

    async def _watchdog(self):
        """Watches and reports peer status changes."""
        try:
            to_remove_interval = int(
                self._max_heartbeats_skipped * self._heartbeat_interval
                + self._heartbeat_interval / 2
            )
            while True:
                failed_peers = set()
                recovered_peers = set()
                degraded_peers = set()
                sleep_interval = self._heartbeat_interval + 5
                # Check all peers for which heartbeats were sent and not responded.
                for identity, timestamp in self._last_heartbeat.items():
                    last_seen = self._known_peers[identity]
                    self.logger.debug(
                        "Watchdog (%s) observing peer %s", self.name, identity
                    )
                    self.logger.debug("Last peer activity: %s.", last_seen)
                    self.logger.debug("Last peer heartbeat sent: %s.", timestamp)
                    self.logger.debug("Since last ping: %d", now() - timestamp)
                    self.logger.debug("Since last seen: %d", now() - last_seen)
                    self.logger.debug(
                        "Until to remove: %d", last_seen + to_remove_interval - now()
                    )
                    self.logger.debug("To remove interval: %d", to_remove_interval)
                    # Message received after heartbeat was sent.
                    # If peer was in degraded state before move it to recovered.
                    if last_seen > timestamp:
                        recovered_peers.add(identity)
                    # Peer is unresponsive.
                    elif now() - last_seen > to_remove_interval:
                        failed_peers.add(identity)
                        continue
                    # Peer failed to answer for 2 heartbeat intervals, entering
                    # degraded state.
                    elif (
                        now() - last_seen > 2 * self._heartbeat_interval
                        and identity not in self._degraded_peers
                    ):
                        degraded_peers.add(identity)

                    till_next_heartbeat = int(
                        timestamp + self._heartbeat_interval - now()
                    )
                    till_remove = int(last_seen + to_remove_interval - now())

                    sleep_interval = min(
                        sleep_interval, till_remove, till_next_heartbeat
                    ) + random.uniform(2, 5)
                self.logger.debug(
                    "Watchdog (%s) adjusting sleep interval: %d",
                    self.name,
                    sleep_interval,
                )

                for recovered_peer in recovered_peers:
                    del self._last_heartbeat[recovered_peer]
                    if recovered_peer in self._degraded_peers:
                        self.logger.debug("Peer %s recovering.", recovered_peer)
                        self._degraded_peers.remove(recovered_peer)
                        await self._status_changed(
                            recovered_peer, PeerStatus.RESPONSIVE
                        )
                for failed_peer in failed_peers:
                    self.logger.debug(
                        "Peer %s entering unresponsive state.", failed_peer
                    )
                    del self._known_peers[failed_peer]
                    del self._last_heartbeat[failed_peer]
                    await self._status_changed(failed_peer, PeerStatus.UNRESPONSIVE)

                for degraded_peer in degraded_peers:
                    self.logger.debug("Peer %s entering degraded state.", degraded_peer)
                    self._degraded_peers.add(degraded_peer)
                    await self._status_changed(degraded_peer, PeerStatus.DEGRADED)
                self.logger.debug(
                    "Watchdog (%s) sleeping for %d seconds.", self.name, sleep_interval
                )
                await asyncio.sleep(sleep_interval)
        except asyncio.CancelledError:
            logger.info("Watchdog shutting down.")
        except:
            logger.exception("Unhandled exception in watchdog.")

    async def start_listening(self):
        """Start listening for messages.

        There are two types of messages: commands and responses.

        When a command is received a command_handler callback is called and
        the response returned by it is sent back to the peer.

        A response is first paired with the sent command and the Event object
        is set so that response is received in the send_command method.
        """
        try:
            while True:
                received = await self._receive_message()
                self.logger.debug(
                    f"Communicator {self.name}: received message: {received}"
                )

                if received is None:
                    self.logger.info(
                        f"Communicator {self.name}: received empty message, closing communicator."
                    )
                    break

                identity, message = received

                # Only add to known peers if the identity is an integer.
                # When executor communicates with listener (sending logs for
                # instance) in uses identity contianing letters. Such should
                # should not be added to known_peers dict since we do not have
                # to monitor them using watchdog.
                with suppress(ValueError):
                    int(identity)
                    if identity not in self._known_peers:
                        self.logger.debug("Adding new peer with identity %s.", identity)
                        self._known_peers[identity] = now()

                if message.message_type is MessageType.HEARTBEAT:
                    # Send response in the background.
                    asyncio.ensure_future(
                        self._send_message(
                            message.respond_heartbeat(), identity=identity
                        )
                    )

                elif message.message_type is MessageType.HEARTBEAT_RESPONSE:
                    if identity in self._last_heartbeat:
                        del self._last_heartbeat[identity]
                    else:
                        self.logger.warning(
                            "The identity %s sent unknown heartbear response.", identity
                        )

                elif message.message_type is MessageType.ENQUIRE:
                    # Send response in the background.
                    asyncio.ensure_future(
                        self._send_message(
                            message.respond(
                                self._message_status(message.message_data).value
                            ),
                            identity=identity,
                        )
                    )

                elif message.message_type is MessageType.COMMAND:
                    if message.uuid not in self._uuids_received:
                        # If final command is received remove the peer from
                        # the list of known peers.
                        if message.command_name == self.FINAL_COMMAND_NAME:
                            if identity in self._known_peers:
                                del self._known_peers[identity]
                            if identity in self._last_heartbeat:
                                del self._last_heartbeat[identity]
                        self._command_queue.append((identity, message))
                        self.logger.debug("New message has arrived")
                        self.logger.debug(
                            f"Number of messages in queue: {len(self._command_queue)}"
                        )

                        self.has_message.set()

                elif message.message_type is MessageType.RESPONSE:
                    # Response from the peer. Match it with the command.
                    response_event = self._uuid_to_event.get(message.uuid, None)
                    if response_event is None:
                        self.logger.error(
                            f"Communicator {self.name}: Can not match response {message} with sent command."
                        )
                    else:
                        response_event.response = message
                        response_event.set()

                else:
                    # Unknown message received. Log it and continue.
                    self.logger.error(
                        f"Communicator {self.name} got unknown message: {message}."
                    )

                self._uuids_received.add(message.uuid)

        except Exception:
            self.logger.exception("Exception while listening for messages.")

        self.logger.info(f"Communicator {self.name} stopped listening for commands.")
        await self.stop_listening()

    async def get_next_message(self) -> Tuple[PeerIdentity, Message]:
        """Get the next message.

        :raises IndexError: if there is no message to process.
        """
        if len(self._command_queue) == 1:
            self.has_message.clear()
        return self._command_queue.popleft()

    async def send_response(
        self, response: Response[MessageDataType], peer_identity: PeerIdentity
    ):
        """Send response."""
        self.logger.debug("Communicator %s: sending response %s", self.name, response)
        await self._send_message(response, peer_identity)

    async def send_command(
        self,
        command: Message[MessageDataType],
        peer_identity: PeerIdentity = b"",
        response_timeout: Optional[int] = 180,
        enquire_timeout: int = 300,
    ) -> Response:
        """Send command and return the response.

        :raises RuntimeError: when response is not received within the
            given timeout.
        """
        command.sequence_number = self.sequence_numbers.setdefault(peer_identity, 1)
        self.sequence_numbers[peer_identity] += 1

        self.logger.debug("Communicator %s: sending command %s.", self.name, command)
        await self._send_message(command, peer_identity)
        response_received_event = EventWithResponse()

        async def enquire(message_uuid):
            """Enquire aboudt message status after enquire_timeout."""
            await asyncio.sleep(enquire_timeout)
            return await self._get_peer_message_status(message_uuid)

        try:
            # Both futures will be canceled in the finally clause.
            enquire_future = asyncio.ensure_future(enquire(command.uuid))
            response_received_future = asyncio.ensure_future(
                response_received_event.wait()
            )

            self._uuid_to_event[command.uuid] = response_received_event
            response = None
            start = time.time()
            done, _ = await asyncio.wait(
                (response_received_future, enquire_future),
                timeout=response_timeout,
                return_when=asyncio.FIRST_COMPLETED,
            )

            if response_received_future in done:
                return response_received_event.response

            elif enquire_future in done:
                if enquire_future.result() == MessageStatus.UNKNOWN.value:
                    raise RuntimeError(
                        f"Communicator {self.name}: response not received by peer."
                    )
            else:
                raise RuntimeError(
                    f"Communicator {self.name}: timeout for response to {command} has expired."
                )

            # Enquire future was successfull, peer is probably just processing the message.
            # Wait for response until timeout.
            if response_timeout is None:
                remaining_timeout = None
            else:
                remaining_timeout = max(0, start + response_timeout - time.time())
            try:
                await asyncio.wait_for(
                    response_received_future, timeout=remaining_timeout
                )
            except asyncio.CancelledError:
                raise RuntimeError(
                    f"Communicator {self.name}: timeout for response to {command} has expired."
                )

            response = response_received_event.response
            self.logger.debug("Received response: %s", response)
        finally:
            enquire_future.cancel()
            response_received_future.cancel()

        return response

    async def stop_listening(self):
        """Stop listening.

        Also stop all running commands.

        This method should only be used when not communicator object is not
        used as a context manager.
        """
        self.logger.debug("Communicator %s stopped listening", self.name)
        self._terminating.set()

    async def terminate(self, reason: str):
        """Stop listening and close the socket.

        Also notify the peer about termination.

        :raises AssertionError: when no listening task is running.
        """
        if self._listening_future is not None:
            with suppress(RuntimeError):
                await self.send_command(
                    Message(MessageType.COMMAND, "terminate", reason)
                )
            await self.stop_listening()
            with suppress(asyncio.CancelledError):
                await self._listening_future
        self.writer.close()


class SocketCommunicator(BaseCommunicator):
    """Handles socket communication."""

    def __init__(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
        name: str,
        logger: logging.Logger,
        peer_status_changed: Optional[
            Callable[[PeerIdentity, PeerStatus], Coroutine]
        ] = None,
    ):
        """Initialize."""
        super().__init__(
            name,
            logger,
            reader,
            writer,
            async_send_data,
            async_receive_data,
            peer_status_changed,
        )


class BaseProtocol:
    """Base protocol class."""

    def __init__(
        self,
        communicator: BaseCommunicator,
        logger: logging.Logger,
    ):
        """Initialize."""
        self.communicator = communicator
        self.logger = logger
        self._should_stop = asyncio.Event()

    async def process_command(
        self, peer_identity: PeerIdentity, received_message: Message
    ):
        """Process single command."""
        command_name = received_message.type_data
        handler_name = "handle_" + command_name
        handler = getattr(self, handler_name, None)
        if handler is None:
            handler = self.default_command_handler
        try:
            response = await handler(received_message, peer_identity)
        except Exception as ex:
            self.logger.exception(
                f"Exception while running command handler '{handler_name}'"
            )
            response = received_message.respond_error(
                f"Exception while running command handler {handler_name}: {ex}"
            )

        try:
            await self.communicator.send_response(response, peer_identity)
        except RuntimeError:
            self.logger.exception(
                "Protocol: error sending response to {received_message}."
            )
            await self._abort_with_error("Error sending response.")

    def post_processing_command(
        self,
        peer_identity: PeerIdentity,
        received_message: Message,
        future: asyncio.Future,
    ):
        """Run optional post-processing command."""
        command_name = received_message.type_data
        handler_name = "post_" + command_name
        handler = getattr(self, handler_name, None)
        if handler is not None:
            asyncio.ensure_future(handler(received_message, peer_identity))

    async def default_command_handler(
        self, message: Message, identity: PeerIdentity
    ) -> Response:
        """Handle the response when no other handler is found for the given command."""
        return message.respond(
            f"No handler for command {message.type_data}.", ResponseStatus.ERROR
        )

    async def handle_terminate(
        self, message: Message, identity: PeerIdentity
    ) -> Response[str]:
        """Handle terminate command."""
        return message.respond_ok("OK")

    async def handle_terminating(
        self, message: Message, identity: PeerIdentity
    ) -> Response[str]:
        """Handle peer termination."""
        return message.respond_ok("OK")

    async def communicate(self):
        """Start communication.

        This coroutine stopss when it is canceled or _should_stop flag is set.
        """
        async with self.communicator:
            try:
                while not self._should_stop.is_set():
                    should_stop_future = asyncio.ensure_future(self._should_stop.wait())
                    communicator_terminating_future = asyncio.ensure_future(
                        self.communicator._terminating.wait()
                    )

                    has_message_future = asyncio.ensure_future(
                        self.communicator.has_message.wait()
                    )

                    self.logger.debug(
                        "Communicate %s: waiting for message.", self.communicator.name
                    )
                    done, pending = await asyncio.wait(
                        (
                            should_stop_future,
                            has_message_future,
                            communicator_terminating_future,
                        ),
                        return_when=asyncio.FIRST_COMPLETED,
                    )
                    self.logger.debug(
                        "Communicate %s got nudged.", self.communicator.name
                    )

                    if has_message_future in done:
                        self.logger.debug(
                            "Communicate %s: got new message.", self.communicator.name
                        )
                        try:
                            (
                                peer_identity,
                                received_command,
                            ) = await self.communicator.get_next_message()
                        except IndexError:
                            # On the next iteration the while loop will stop.
                            self.logger.exception(
                                "Communicator: error receiving next message."
                            )
                            await self._abort_with_error("Error receiving next message")
                        else:
                            # Run process command in the background so the main
                            # process can answer other queries.
                            processing_future = asyncio.ensure_future(
                                self.process_command(peer_identity, received_command)
                            )
                            callback = functools.partial(
                                self.post_processing_command,
                                peer_identity,
                                received_command,
                            )
                            processing_future.add_done_callback(callback)
                    elif communicator_terminating_future in done:
                        # Communicator is closed. Stop.
                        self.logger.debug(
                            "Communicator %s stoped", self.communicator.name
                        )
                        self.stop_communicate()

                    for future in pending:
                        self.logger.debug(
                            "Communicate (%s): cancelling future",
                            self.communicator.name,
                        )
                        future.cancel()
                        self.logger.debug(
                            "Communicate (%s): future cancelled", self.communicator.name
                        )
            except:
                self.logger.exception("Exception while running communicate.")
                await self._abort_with_error("Exception while running communicate.")
        self.logger.debug(
            "Communicator %s stopped communicating", self.communicator.name
        )

    def stop_communicate(self):
        """Stop communicating."""
        self.logger.debug(
            "Protocol using %s: stop communicate set.", self.communicator.name
        )
        self._should_stop.set()

    async def _abort_with_error(self, error: str):
        """Log error and continue by default."""
        self.logger.error(error)



# pylint: disable=missing-docstring
from resolwe.flow.models import Data, DescriptorSchema, Process
from resolwe.test import TestCase


class DescriptorTestCase(TestCase):
    def setUp(self):
        super().setUp()

        self.process = Process.objects.create(
            name="Dummy process", contributor=self.contributor
        )

        self.descriptor_schema = DescriptorSchema.objects.create(
            name="Descriptor schema",
            contributor=self.contributor,
            schema=[
                {
                    "name": "test_field",
                    "type": "basic:string:",
                    "default": "default value",
                }
            ],
        )

    def test_default_values(self):
        data = Data.objects.create(
            name="Data object",
            contributor=self.contributor,
            process=self.process,
            descriptor_schema=self.descriptor_schema,
        )
        self.assertEqual(data.descriptor["test_field"], "default value")

        data = Data.objects.create(
            name="Data object 2",
            contributor=self.contributor,
            process=self.process,
            descriptor_schema=self.descriptor_schema,
            descriptor={"test_field": "changed value"},
        )
        self.assertEqual(data.descriptor["test_field"], "changed value")



"""Data transfer between connectors."""
import concurrent.futures
import logging
from contextlib import suppress
from functools import partial
from pathlib import Path
from time import sleep
from typing import TYPE_CHECKING, Iterable, List, Optional, Union

import wrapt
from requests.exceptions import ConnectionError as RequestsConnectionError
from requests.exceptions import ReadTimeout

from .baseconnector import BaseStorageConnector
from .circular_buffer import CircularBuffer
from .exceptions import DataTransferError
from .utils import paralelize

if TYPE_CHECKING:
    from os import PathLike

try:
    from google.api_core.exceptions import ServiceUnavailable
    from google.resumable_media.common import DataCorruption

    gcs_exceptions = [DataCorruption, ServiceUnavailable]
except ModuleNotFoundError:
    gcs_exceptions = []


logger = logging.getLogger(__name__)
ERROR_MAX_RETRIES = 3
ERROR_TIMEOUT = 5  # In seconds.
transfer_exceptions = tuple(
    gcs_exceptions + [DataTransferError] + [RequestsConnectionError, ReadTimeout]
)


@wrapt.decorator
def retry_on_transfer_error(wrapped, instance, args, kwargs):
    """Retry on tranfser error."""
    for _ in range(ERROR_MAX_RETRIES):
        try:
            return wrapped(*args, **kwargs)
        except transfer_exceptions as err:
            connection_err = err
            sleep(ERROR_TIMEOUT)

    raise connection_err


class Transfer:
    """Transfer data between two storage connectors using in-memory buffer."""

    def __init__(
        self,
        from_connector: "BaseStorageConnector",
        to_connector: "BaseStorageConnector",
    ):
        """Initialize transfer object."""
        self.from_connector = from_connector
        self.to_connector = to_connector

    def pre_processing(
        self, url: Union[str, Path], objects: Optional[List[dict]] = None
    ):
        """Notify connectors that transfer is about to start.

        The connector is allowed to change names of the objects that are to be
        transfered. This allows us to do some pre-processing, like zipping all
        files into one and transfering that one.

        :param url: base url for file transfer.

        :param objects: list of objects to be transfered, their paths are
            relative with respect to the url.
        """
        objects_to_transfer = self.from_connector.before_get(objects, url)
        self.to_connector.before_push(objects_to_transfer, url)
        return objects_to_transfer

    def post_processing(
        self, url: Union[str, Path], objects: Optional[List[dict]] = None
    ):
        """Notify connectors that transfer is complete.

        :param url: base url for file transfer.

        :param objects: the list ob objects that was actually transfered.The
            paths are relative with respect to the url.
        """
        self.from_connector.after_get(objects, url)
        objects_stored = self.to_connector.after_push(objects, url)
        return objects_stored

    def transfer_objects(
        self, url: Union[str, Path], objects: List[dict], max_threads: int = 10
    ) -> Optional[List[dict]]:
        """Transfer objects under the given URL.

        Objects are read from from_connector and copied to to_connector.

        :param url: the given URL to transfer from/to.

        :param objects: the list of objects to transfer. Each object is
            represented with the dictionary containing at least keys "path",
            "size", "md5", "crc32c", "awss3etag", "chunk_size".
            All values for key "path" must be relative with respect to the
            argument url.

        :returns: the list of objects that were stored in the to_connector if
            it is different that argument objects or None.
        """
        # Pre-processing.
        try:
            objects_to_transfer = self.pre_processing(url, objects)
        except Exception:
            logger.exception(
                "Error in pre-processing while transfering data from url {}".format(url)
            )
            raise DataTransferError()

        url = Path(url)

        futures = paralelize(
            objects=objects_to_transfer,
            worker=partial(self.transfer_chunk, url),
            max_threads=max_threads,
        )

        # Check future results. This wil re-raise any exception raised in
        # _transfer_chunk.
        if not all(future.result() for future in futures):
            raise DataTransferError()

        # Post-processing.
        try:
            objects_stored = self.post_processing(url, objects_to_transfer)
        except Exception:
            logger.exception(
                "Error in post-processing while transfering data from url {}".format(
                    url
                )
            )
            raise DataTransferError()

        return None if objects_stored is objects else objects_stored

    def transfer_chunk(self, url: Path, objects: Iterable[dict]) -> bool:
        """Transfer a single chunk of objects.

        When objects have properties `from_base_url` and `to_base_url` they
        override the `url` argument.

        :raises DataTransferError: on failure.
        :returns: True on success.
        """
        to_connector = self.to_connector.duplicate()
        from_connector = self.from_connector.duplicate()
        for entry in objects:
            # Do not transfer directories.
            if not entry["path"].endswith("/"):
                if not self.transfer(
                    entry.get("from_base_url", url),
                    entry,
                    entry.get("to_base_url", url),
                    Path(entry["path"]),
                    from_connector,
                    to_connector,
                ):
                    raise DataTransferError()
        return True

    @retry_on_transfer_error
    def transfer(
        self,
        from_base_url: Union[str, Path],
        object_: dict,
        to_base_url: Union[str, Path],
        to_url: "PathLike[str]",
        from_connector: "BaseStorageConnector" = None,
        to_connector: "BaseStorageConnector" = None,
    ) -> bool:
        """Transfer single object between two storage connectors.

        :param from_base_url: base url on from_connector.

        :param object_: object to transfer. It must be a dictionary containing
            at least keys "path", "md5", "crc32c", "size" and "awss3etag". It
            can also contain key "chunk_size" that specifies a custom
            chunk_size to use for upload / download.

        :param to_base_url: base url on to_connector.

        :param to_url: where to copy object. It is relative with respect to the
            argument to_base_url.

        :param from_connector: from connector, defaults to None. If None
            duplicate of from_connector from the Transfer class instance is
            used.

        :param to_connector: to connector, defaults to None. If None
            duplicate of to_connector from the Transfer class instance is
            used.

        :raises DataTransferError: on failure.

        :returns: True on success.
        """
        to_base_url = Path(to_base_url)
        chunk_size = object_.get("chunk_size", BaseStorageConnector.CHUNK_SIZE)
        # Duplicate connectors for thread safety.
        to_connector = to_connector or self.to_connector.duplicate()
        from_connector = from_connector or self.from_connector.duplicate()

        from_url = Path(from_base_url) / object_["path"]
        hashes = {type_: object_[type_] for type_ in ["md5", "crc32c", "awss3etag"]}

        skip_final_hash_check = (
            from_connector.get_ensures_data_integrity
            and to_connector.put_ensures_data_integrity
        )
        if skip_final_hash_check:
            # When final check is skipped make sure that the input connector
            # hash equals to the hash given by the _object (usually read from
            # the database).
            hash_to_check = next(
                hash for hash in from_connector.supported_hash if hash in hashes.keys()
            )
            from_connector_hash = from_connector.get_hash(from_url, hash_to_check)
            expected_hash = object_[hash_to_check]
            if expected_hash != from_connector_hash:
                raise DataTransferError(
                    f"Connector {from_connector} has {from_connector_hash} stored  "
                    f"as {from_connector_hash} hash for object "
                    f"{from_url}, expected {expected_hash}."
                )

        common_hash_type = next(
            e for e in to_connector.supported_hash if e in hashes.keys()
        )
        from_hash = hashes[common_hash_type]

        # Check if file already exist and has the right hash.
        to_hash = to_connector.get_hash(to_base_url / to_url, common_hash_type)
        if from_hash == to_hash:
            # Object exists and has the right hash.
            logger.debug(
                "From: {}:{}".format(from_connector.name, from_url)
                + " to: {}:{}".format(to_connector.name, to_base_url / to_url)
                + " object exists with right hash, skipping."
            )
            return True

        # When object can be open directly as stream do it.
        if from_connector.can_open_stream:
            stream = from_connector.open_stream(from_url, "rb")
            to_connector.push(stream, to_base_url / to_url, chunk_size=chunk_size)
            stream.close()

        elif to_connector.can_open_stream:
            stream = to_connector.open_stream(to_base_url / to_url, "wb")
            from_connector.get(from_url, stream, chunk_size=chunk_size)
            stream.close()
        # Otherwise create out own stream and use threads to transfer data.
        else:

            def future_done(stream_to_close, future):
                stream_to_close.close()
                if future.exception() is not None:
                    executor.shutdown(wait=False)

            data_stream = CircularBuffer(
                buffer_size=min(200 * 1024 * 1024, object_["size"])
            )
            with concurrent.futures.ThreadPoolExecutor() as executor:
                download_task = executor.submit(
                    from_connector.get,
                    from_url,
                    data_stream,
                    chunk_size=chunk_size,
                )
                upload_task = executor.submit(
                    to_connector.push,
                    data_stream,
                    to_base_url / to_url,
                    chunk_size=chunk_size,
                )
                download_task.add_done_callback(partial(future_done, data_stream))
                futures = (download_task, upload_task)

            # Re-raise possible exception as DataTransferError.
            if any(f.exception() is not None for f in futures):
                # Log exceptions in threads to preserve original stack trace.
                for f in futures:
                    try:
                        f.result()
                    except Exception:
                        logger.exception("Exception occured while transfering data")

                # Delete transfered data.
                with suppress(Exception):
                    to_connector.delete(to_base_url, [to_url])

                # Re-raise exception.
                ex = [f.exception() for f in futures if f.exception() is not None]
                messages = [str(e) for e in ex]
                raise DataTransferError("\n\n".join(messages))

        # Check hash of the uploaded object.
        if not skip_final_hash_check:
            to_hash = to_connector.get_hash(to_base_url / to_url, common_hash_type)
            if from_hash != to_hash:
                with suppress(Exception):
                    to_connector.delete(to_base_url, [to_url])
                raise DataTransferError(
                    f"Hash {common_hash_type} does not match while transfering "
                    f"{from_url} -> {to_base_url/to_url}: using hash type "
                    f"{common_hash_type}: expected {from_hash}, got {to_hash}."
                )

        # Store computed hashes as metadata for later use.
        to_connector.set_hashes(to_base_url / to_url, hashes)

        return True
